import sys
from typing import Union, Dict, List, Tuple, Optional

import numpy as np
import scipy.signal
import seaborn as sns
import matplotlib.pyplot as plt

import sanpy

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


def old_getEddLines(ba):
    """Get lines representing linear fit of EDD rate.

    Args:
        ba (bAnalysis): bAnalysis object
    """
    logger.info(ba)

    x = []
    y = []
    if ba is None or ba.numSpikes == 0:
        return x, y

    # these are getting for current sweep
    preLinearFitPnt0 = ba.getStat("preLinearFitPnt0")
    preLinearFitSec0 = [ba.pnt2Sec_(x) for x in preLinearFitPnt0]
    preLinearFitVal0 = ba.getStat("preLinearFitVal0")

    preLinearFitPnt1 = ba.getStat("preLinearFitPnt1")
    preLinearFitSec1 = [ba.pnt2Sec_(x) for x in preLinearFitPnt1]
    preLinearFitVal1 = ba.getStat("preLinearFitVal1")

    thisNumSpikes = len(preLinearFitPnt0)
    # for idx, spike in enumerate(range(ba.numSpikes)):
    for idx, spike in enumerate(range(thisNumSpikes)):
        dx = preLinearFitSec1[idx] - preLinearFitSec0[idx]
        dy = preLinearFitVal1[idx] - preLinearFitVal0[idx]

        lineLength = 8  # TODO: make this a function of spike frequency?

        x.append(preLinearFitSec0[idx])
        x.append(preLinearFitSec1[idx] + lineLength * dx)
        x.append(np.nan)

        y.append(preLinearFitVal0[idx])
        y.append(preLinearFitVal1[idx] + lineLength * dy)
        y.append(np.nan)

        # print('here')

    return x, y


def old_getHalfWidths(ba):
    """Get lines representing half-widhts (AP Dur).

    Args:
        ba (bAnalysis): bAnalysis object

    Returns:
        x,y (enum):
    """
    # defer until we know how many half-widths 20/50/80
    x = []
    y = []
    numPerSpike = 3  # rise/fall/nan
    # numSpikes = ba.numSpikes
    xyIdx = 0
    spikeDictionaries = ba.getSpikeDictionaries()  # for current sweep
    numSpikes = len(spikeDictionaries)
    # for idx, spike in enumerate(ba.spikeDict):
    for idx, spike in enumerate(spikeDictionaries):
        if idx == 0:
            # make x/y from first spike using halfHeights = [20,50,80]
            halfHeights = spike[
                "halfHeights"
            ]  # will be same for all spike, like [20, 50, 80]
            numHalfHeights = len(halfHeights)
            # *numHalfHeights to account for rise/fall + padding nan
            x = [np.nan] * (numSpikes * numHalfHeights * numPerSpike)
            y = [np.nan] * (numSpikes * numHalfHeights * numPerSpike)
            # print('  len(x):', len(x), 'numHalfHeights:', numHalfHeights, 'numSpikes:', numSpikes, 'halfHeights:', halfHeights)

        for idx2, width in enumerate(spike["widths"]):
            halfHeight = width["halfHeight"]  # [20,50,80]
            risingPnt = width["risingPnt"]
            risingVal = width["risingVal"]
            fallingPnt = width["fallingPnt"]
            fallingVal = width["fallingVal"]

            if risingPnt is None or fallingPnt is None:
                # half-height was not detected
                continue

            risingSec = ba.pnt2Sec_(risingPnt)
            fallingSec = ba.pnt2Sec_(fallingPnt)

            x[xyIdx] = risingSec
            x[xyIdx + 1] = fallingSec
            x[xyIdx + 2] = np.nan
            # y
            y[xyIdx] = fallingVal  # risingVal, to make line horizontal
            y[xyIdx + 1] = fallingVal
            y[xyIdx + 2] = np.nan

            # each spike has 3x pnts: rise/fall/nan
            xyIdx += numPerSpike  # accounts for rising/falling/nan
        # end for width
    # end for spike
    return x, y


class bAnalysisPlot:
    """
    Class to plot results of [sanpy.bAnalysis][sanpy.bAnalysis] spike detection.

    """

    def __init__(self, ba=None):
        """
        Args:
            ba: [sanpy.bAnalysis][sanpy.bAnalysis] object
        """
        self._ba = ba

    @property
    def ba(self):
        """Get underlying bAnalysis object."""
        return self._ba

    def getDefaultPlotStyle(self):
        """Get dictionary with default plot style."""
        d = {
            "linewidth": 1,
            "color": "k",
            "width": 9,
            "height": 3,
        }
        return d.copy()

    def _makeFig(self, plotStyle=None):
        if plotStyle is None:
            plotStyle = self.getDefaultPlotStyle()

        grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

        width = plotStyle["width"]
        height = plotStyle["height"]
        fig = plt.figure(figsize=(width, height))
        ax = fig.add_subplot(grid[0, 0:])  # Vm, entire sweep

        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)

        return fig, ax

    def plotRaw(self, plotStyle=None, ax=None):
        """
        Plot raw recording

        Args:
            plotStye (float):
            ax (xxx):
        """

        if plotStyle is None:
            plotStyle = self.getDefaultPlotStyle()

        if ax is None:
            _fig, _ax = self._makeFig()
            _fig.suptitle(f"{self._ba.fileLoader.filepath}")
        else:
            _ax = ax

        color = plotStyle["color"]
        linewidth = plotStyle["linewidth"]
        sweepX = self.ba.fileLoader.sweepX
        sweepY = self.ba.fileLoader.sweepY

        _ax.plot(
            sweepX, sweepY, "-", c=color, linewidth=linewidth
        )  # fmt = '[marker][line][color]'

        xUnits = self.ba.fileLoader.get_xUnits()
        yUnits = self.ba.fileLoader.get_yUnits()
        _ax.set_xlabel(xUnits)
        _ax.set_ylabel(yUnits)

        return _ax

    def plotDerivAndRaw(self):
        """
        Plot both Vm and the derivative of Vm (dV/dt).

        Args:
            fig (matplotlib.pyplot.figure): An existing figure to plot to.
                see: https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html

        Return:
            fig and axs
        """

        #
        # make a 2-panel figure
        grid = plt.GridSpec(2, 1, wspace=0.2, hspace=0.4)
        fig = plt.figure(figsize=(10, 8))
        ax1 = fig.add_subplot(grid[0, 0])
        ax2 = fig.add_subplot(grid[1, 0], sharex=ax1)
        ax1.spines["right"].set_visible(False)
        ax1.spines["top"].set_visible(False)
        ax2.spines["right"].set_visible(False)
        ax2.spines["top"].set_visible(False)

        self.plotRaw(ax=ax1)

        sweepX = self.ba.fileLoader.sweepX
        filteredDeriv = self.ba.fileLoader.filteredDeriv
        ax2.plot(sweepX, filteredDeriv)

        ax2.set_ylabel("dV/dt")
        # ax2.set_xlabel('Seconds')

        return fig, [ax1, ax2]

    def plotSpikes(
        self,
        plotThreshold=False,
        plotPeak=False,
        plotStyle=None,
        hue: Optional[str] = "condition",
        markerSize: Optional[int] = 4,
        ax=None,
    ):
        """Plot Vm with spike analysis overlaid as symbols

        Args:
            plotThreshold
            plotPeak
            plotStyle (dict): xxx
            markerSize
            ax (xxx): If specified will plot into a MatPlotLib axes

        Returns
            fig
            ax
        """

        legend = False

        if plotStyle is None:
            plotStyle = self.getDefaultPlotStyle()

        if ax is None:
            fig, ax = self._makeFig()
            fig.suptitle(self.ba.getFileName())
        # plot vm
        self.plotRaw(ax=ax)

        # plot spike times
        if plotThreshold:
            thresholdSec = self.ba.getStat("thresholdSec")  # x
            thresholdVal = self.ba.getStat("thresholdVal")  # y

            ax.plot(thresholdSec, thresholdVal, "pr", markersize=markerSize)
            df = self.ba.spikeDict.asDataFrame()  # regenerates, can be expensive
            # logger.info(f'df hue unique is: {df[hue].unique()}')
            # sns.scatterplot(x='thresholdSec', y='thresholdVal', hue=hue, data=df, ax=ax, legend=legend)

        # plot the peak
        if plotPeak:
            peakPnt = self.ba.getStat("peakPnt")
            # don't use this for Ca++ concentration
            # peakVal = self.ba.getStat('peakVal')
            peakVal = self.ba.fileLoader.sweepY[peakPnt]
            peakSec = [self.ba.fileLoader.pnt2Sec_(x) for x in peakPnt]
            ax.plot(peakSec, peakVal, "oc", markersize=markerSize)

        xUnits = self.ba.fileLoader.get_xUnits()
        yUnits = self.ba.fileLoader.get_yUnits()
        ax.set_xlabel(xUnits)
        ax.set_ylabel(yUnits)

        return ax

    def plotStat(self, xStat: str, yStat: str, hue: Optional[str] = None, ax=None):
        # ax : Optional["matplotlib.axes._subplots.AxesSubplot"] = None):

        legend = False

        if ax is None:
            fig, ax = self._makeFig()
            fig.suptitle(self.ba.getFileName())
        print(type(ax))

        df = self.ba.spikeDict.asDataFrame()  # regenerates, can be expensive
        sns.scatterplot(x=xStat, y=yStat, hue=hue, data=df, ax=ax, legend=legend)

    def plotTimeSeries(ba, stat, halfWidthIdx=0, ax=None):
        """Plot a given spike parameter."""
        if stat == "peak":
            yStatName = "peakVal"
            yStatLabel = "Spike Peak (mV)"
        if stat == "preMin":
            yStatName = "preMinVal"
            yStatLabel = "Pre Min (mV)"
        if stat == "halfWidth":
            yStatName = "widthPnts"
            yStatLabel = "Spike Half Width (ms)"

        #
        # pull
        statX = []
        statVal = []
        for i, spike in enumerate(ba.spikeDict):
            if i == 0 or i == len(ba.spikeTimes) - 1:
                continue
            else:
                statX.append(spike["peakSec"])
                if stat == "halfWidth":
                    statVal.append(spike["widths"][halfWidthIdx]["widthMs"])
                else:
                    statVal.append(spike[yStatName])

        #
        # plot
        if ax is None:
            grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(grid[0, 0:])  # Vm, entire sweep

        ax.plot(statX, statVal, "o-k")

        ax.set_ylabel(yStatLabel)
        ax.set_xlabel("Time (sec)")

        return statVal

    def plotISI(ba, ax=None):
        """Plot the inter-spike-interval (sec) between each spike threshold"""
        #
        # pull
        spikeTimes_sec = [x / ba.dataPointsPerMs / 1000 for x in ba.spikeTimes]
        isi = np.diff(spikeTimes_sec)
        isi_x = spikeTimes_sec[0:-1]

        #
        # plot
        if ax is None:
            grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(grid[0, 0:])  # Vm, entire sweep

        ax.plot(isi_x, isi, "o-k")

        ax.set_ylabel("Inter-Spike-Interval (sec)")
        ax.set_xlabel("Time (sec)")

    def plotClips(
        self, plotType="Raw", preClipWidth_ms=None, postClipWidth_ms=None, ax=None
    ):
        """Plot clips of all detected spikes

        Clips are created in self.spikeDetect() and default to clipWidth_ms = 100 ms

        Args:
            plotType (str): From [Raw, RawPlusMean, Mean, SD, Var]
            #plotVariance (bool): If tru, plot variance. Otherwise plot all raw clips (black) with mean (red)

        Returns:
            xPlot
            yPlot
        """

        """
        if ax is None:
            grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(grid[0, 0:]) #Vm, entire sweep
        """

        if self.ba.numSpikes == 0:
            return None, None

        if ax is None:
            fig, ax = self._makeFig()
            fig.suptitle(self.ba.getFileName())

        startSec, stopSec = None, None
        selectedSpikeList = []
        # preClipWidth_ms = 200
        # postClipWidth_ms = 1000

        # Leave this none, if we pass none, ba.getSpikeClipsWill take care of this
        """
        if preClipWidth_ms is None:
            preClipWidth_ms = self.ba.detectionClass['preSpikeClipWidth_ms']
        if postClipWidth_ms is None:
            postClipWidth_ms = self.ba.detectionClass['postSpikeClipWidth_ms']
        """
        sweepNumber = 0
        theseClips, theseClips_x, meanClip = self.ba.getSpikeClips(
            startSec,
            stopSec,
            spikeSelection=selectedSpikeList,
            preSpikeClipWidth_ms=preClipWidth_ms,
            postSpikeClipWidth_ms=postClipWidth_ms,
            sweepNumber=sweepNumber,
        )
        numClips = len(theseClips)

        # convert clips to 2d ndarray ???
        xTmp = np.array(theseClips_x)
        # xTmp /= self.ba.dataPointsPerMs * 1000  # pnt to seconds
        xTmp /= 1000  # ms to seconds
        yTmp = np.array(theseClips)  # mV

        # TODO: plot each clip as different color
        # this will show variation in sequential clips (for Ca++ imaging they are decreasing)
        # print(xTmp.shape)  # ( e.g. (9,306)
        # sys.exit(1)

        # plot variance
        if plotType == "Mean":
            xPlot = np.nanmean(xTmp, axis=0)  # xTmp is in ms
            yPlot = np.nanmean(yTmp, axis=0)
            ax.plot(xPlot, yPlot, "-k", linewidth=1)
            ax.set_ylabel("Mean")
            ax.set_xlabel("Time (sec)")
        elif plotType == "Var":
            xPlot = np.nanmean(xTmp, axis=0)  # xTmp is in ms
            yPlot = np.nanvar(yTmp, axis=0)
            ax.plot(xPlot, yPlot, "-k", linewidth=1)
            ax.set_ylabel("Variance")
            ax.set_xlabel("Time (sec)")
        elif plotType == "SD":
            xPlot = np.nanmean(xTmp, axis=0)  # xTmp is in ms
            yPlot = np.nanstd(yTmp, axis=0)
            ax.plot(xPlot, yPlot, "-k", linewidth=1)
            ax.set_ylabel("STD")
            ax.set_xlabel("Time (sec)")
        elif plotType in ["Raw", "RawPlusMean"]:
            # plot raw
            # logger.info('PLOTTING RAW')

            cmap = plt.get_cmap("jet")
            colors = [cmap(i) for i in np.linspace(0, 1, numClips)]

            for i in range(numClips):
                color = colors[i]
                xPlot = xTmp[i, :]
                yPlot = yTmp[i, :]
                # I want different colors here
                # ax.plot(xPlot, yPlot, '-k', linewidth=0.5)
                # WHY IS THIS NOT WORKING !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                # logger.info(f'!!!!!!!!!!!!!!!!!!!!!! plotting color g {plotType} {color}')
                # ax.plot(xPlot, yPlot, '-g', linewidth=0.5, color='g')
                ax.plot(xPlot, yPlot, "-", label=f"{i}", color=color, linewidth=0.5)

            yLabel = self.ba._sweepLabelY
            ax.set_ylabel(yLabel)
            ax.set_xlabel("Time (sec)")

            # plot mean
            if plotType == "RawPlusMean":
                xMeanClip = np.nanmean(xTmp, axis=0)  # xTmp is in ms
                yMeanClip = np.nanmean(yTmp, axis=0)
                ax.plot(xMeanClip, yMeanClip, "-r", linewidth=1)

            # show legend for each raw race 0,1,2,3,...
            ax.legend(loc="best")

        else:
            logger.error(f"Did not understand plot type: {plotType}")

        #
        return xPlot, yPlot

    def plotPhasePlot(self, oneSpikeNumber=None, ax=None):
        if ax is None:
            grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(grid[0, 0:])  # Vm, entire sweep

        filteredClip = scipy.signal.medfilt(self.spikeClips[oneSpikeNumber], 3)
        dvdt = np.diff(filteredClip)
        # add an initial point so it is the same length as raw data in abf.sweepY
        dvdt = np.concatenate(([0], dvdt))
        (line,) = ax.plot(filteredClip, dvdt, "y")

        ax.set_ylabel("filtered dV/dt")
        ax.set_xlabel("filtered Vm (mV)")

        return line


def test_plot(path):
    print("=== test_plot() path:", path)
    ba = sanpy.bAnalysis(path)

    # detect
    dDict = sanpy.bAnalysis.getDefaultDetection()
    dDict["dvdThreshold"] = 50
    ba.spikeDetect(dDict)

    # plot
    bp = sanpy.bAnalysisPlot(ba)

    fig = bp.plotDerivAndRaw()

    fig = bp.plotSpikes()

    plt.show()


if __name__ == "__main__":
    path = "data/19114001.abf"
    test_plot(path)

    # TODO: check if error

    path = "data/19114001.csv"
    test_plot(path)



# created 20210212
import os, math

import numpy as np
import pandas as pd

from PyQt5 import QtCore, QtGui, QtWidgets

import sanpy.analysisDir
import sanpy.bDetection

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


def old_loadDatabase(path):
    """
    Load a csv/xls/xlsx file as a pandas dataframe.

    Args:
        path (Str): full path to .csv file generated with reanalyze.py

    Returns:
        pandas dataframe
    """
    masterDf = None
    if path is None:
        pass
    elif not os.path.isfile(path):
        print(f'error: bUtil.loadDatabase() did not find file: "{path}"')
    elif path.endswith(".csv"):
        masterDf = pd.read_csv(
            path, header=0, index_col=False
        )  # , dtype={'ABF File': str})
    elif path.endswith(".xls"):
        masterDf = pd.read_excel(path, header=0)  # , dtype={'ABF File': str})
    elif path.endswith(".xlsx"):
        masterDf = pd.read_excel(
            path, header=0, engine="openpyxl"
        )  # , dtype={'ABF File': str})
    else:
        print("error: file type not supported. Expecting csv/xls/xlsx. Path:", path)
    #
    return masterDf


def printDict(d, withType=False):
    for k, v in d.items():
        if withType:
            print(f"  {k}: {v} {type(v)}")
        else:
            print(f"  {k}: {v}")


class myTableView(QtWidgets.QTableView):
    """Table view to display list of files.

    TODO: Try and implement the first column (filename) as a frozen column.

    See: https://doc.qt.io/qt-5/qtwidgets-itemviews-frozencolumn-example.html
    """

    signalDuplicateRow = QtCore.pyqtSignal(object)  # row index
    signalDeleteRow = QtCore.pyqtSignal(object)  # row index
    # signalRefreshTabe = QtCore.pyqtSignal(object) # row index
    signalCopyTable = QtCore.pyqtSignal()
    signalFindNewFiles = QtCore.pyqtSignal()
    signalSaveFileTable = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        """ """
        super(myTableView, self).__init__(parent)

        self.doIncludeCheckbox = False  # todo: turn this on
        # need a local reference to delegate else 'segmentation fault'
        # self.keepCheckBoxDelegate = myCheckBoxDelegate(None)

        # self.setFont(QtGui.QFont('Arial', 10))

        self.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.setSelectionBehavior(QtWidgets.QTableView.SelectRows)

        self.setEditTriggers(
            QtWidgets.QAbstractItemView.NoEditTriggers
            | QtWidgets.QAbstractItemView.DoubleClicked
        )

        # removed mar 26 2023
        """
        rowHeight = 11
        fnt = self.font()
        fnt.setPointSize(rowHeight)
        self.setFont(fnt)
        self.verticalHeader().setDefaultSectionSize(rowHeight)
        """

        """
        p = self.palette()
        color1 = QtGui.QColor('#dddddd')
        color2 = QtGui.QColor('#ffffff')
        p.setColor(QtGui.QPalette.Base, color1)
        p.setColor(QtGui.QPalette.AlternateBase, color2)
        self.setPalette(p)
        self.setAlternatingRowColors(True)
        """

    # def keyPressEvent(self, event): #Reimplement the event here, in your case, do nothing
    #    return

    def mySetModel(self, model):
        """
        Set the model. Needed so we can show/hide columns
        """
        self.setModel(model)

        # print('---trying to hide columns')
        # self.hideColumn(1)
        # self.setColumnHidden(1,True);
        # self.setColumnHidden(1, True)
        # self.horizontalHeader().hideSection(1)

    """
    # trying to use this to remove tooltip when it comes up as empty ''
    def viewportEvent(self, event):
        logger.info('')
        return True
    """

    def contextMenuEvent(self, event):
        """Handle right mouse click"""
        contextMenu = QtWidgets.QMenu(self)

        duplicateRow = contextMenu.addAction("Duplicate Row")
        contextMenu.addSeparator()

        # deleteRow = contextMenu.addAction("Delete Row")
        # contextMenu.addSeparator()

        copyTable = contextMenu.addAction("Copy Table")
        contextMenu.addSeparator()

        findNewFiles = contextMenu.addAction("Sync With Folder")
        contextMenu.addSeparator()

        saveTable = contextMenu.addAction("Save Table")

        #
        action = contextMenu.exec_(self.mapToGlobal(event.pos()))
        # logger.info(f'  action:{action}')
        if action == duplicateRow:
            # print('  todo: duplicateRow')
            tmp = self.selectedIndexes()
            if len(tmp) > 0:
                selectedRow = tmp[0].row()
                self.signalDuplicateRow.emit(selectedRow)

        # elif action == deleteRow:
        #     #print('  todo: deleteRow')
        #     tmp = self.selectedIndexes()
        #     if len(tmp)>0:
        #         selectedRow = tmp[0].row()
        #         self.signalDeleteRow.emit(selectedRow)
        #     else:
        #         logger.warning('no selection?')

        elif action == copyTable:
            # print('  todo: copyTable')
            self.signalCopyTable.emit()

        elif action == findNewFiles:
            # print('  todo: findNewFiles')
            self.signalFindNewFiles.emit()

        elif action == saveTable:
            # print('  todo: saveTable')
            self.signalSaveFileTable.emit()

        else:
            logger.warning(f'action not taken "{action}"')


class pandasModel(QtCore.QAbstractTableModel):
    signalMyDataChanged = QtCore.pyqtSignal(object, object, object)
    """Emit on user editing a cell."""

    def __init__(self, data):
        """
        Data model for a pandas dataframe or sanpy.analysisDir.

        Args:
            data (dataframe or analysisDir): pandas dataframe or analysisDir
        """
        QtCore.QAbstractTableModel.__init__(self)

        self.isDirty = False

        # data is either DataFrame or analysisDir
        self.isAnalysisDir = False
        if isinstance(data, pd.core.frame.DataFrame):
            self.isAnalysisDir = False
        elif isinstance(data, sanpy.analysisDir):
            self.isAnalysisDir = True
        else:
            logger.error("Expecting data in (DataFrame, sanpy.analysisDir)")
        self._data = data

        # self.setSortingEnabled(True)

    """
    def modelReset(self):
        print('modelReset()')
    """

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parnet=None):
        return self._data.shape[1]

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.ToolTipRole:
                # removed
                # swapped sanpy.bDetection.defaultDetection to a class
                # do not want to instantiate every time
                """
                # get default value from bAnalysis
                defaultDetection = sanpy.bDetection.defaultDetection
                columnName = self._data.columns[index.column()]
                toolTip = QtCore.QVariant()  # empty tooltip
                try:
                    toolTip = str(defaultDetection[columnName]['defaultValue'])
                    toolTip += ': ' + defaultDetection[columnName]['description']
                except (KeyError):
                    pass
                return toolTip
                """
            elif role in [QtCore.Qt.DisplayRole, QtCore.Qt.EditRole]:
                columnName = self._data.columns[index.column()]
                if self.isAnalysisDir and columnName == "I":
                    return ""

                # don't get col from index, get from name
                # retVal = self._data.iloc[index.row(), index.column()]
                # retVal = self._data.loc[index.row(), columnName]
                realRow = self._data.index[index.row()]
                # retVal = self._data.iloc[realRow, index.column()]
                retVal = self._data.loc[realRow, columnName]
                if isinstance(retVal, np.float64):
                    retVal = float(retVal)
                elif isinstance(retVal, np.int64):
                    retVal = int(retVal)
                elif isinstance(retVal, str) and retVal == "nan":
                    retVal = ""

                if isinstance(retVal, float) and math.isnan(retVal):
                    # don't show 'nan' in table
                    retVal = ""
                return retVal

            elif role == QtCore.Qt.CheckStateRole:
                columnName = self._data.columns[index.column()]
                realRow = self._data.index[index.row()]
                # retVal = self._data.iloc[index.row(), index.column()]
                retVal = self._data.loc[realRow, columnName]
                if columnName == "I":
                    if retVal:
                        return QtCore.Qt.Checked
                    else:
                        return QtCore.Qt.Unchecked
                return QtCore.QVariant()

            elif role == QtCore.Qt.FontRole:
                realRow = self._data.index[index.row()]
                columnName = self._data.columns[index.column()]
                if columnName == "L":
                    if self._data.isLoaded(realRow):  # or self._data.isSaved(realRow):
                        return QtCore.QVariant(QtGui.QFont("Arial", pointSize=32))
                elif columnName == "A":
                    if self._data.isAnalyzed(
                        realRow
                    ):  # or self._data.isSaved(realRow):
                        return QtCore.QVariant(QtGui.QFont("Arial", pointSize=32))
                elif columnName == "S":
                    if self._data.isSaved(realRow):
                        return QtCore.QVariant(QtGui.QFont("Arial", pointSize=32))
                return QtCore.QVariant()
            elif role == QtCore.Qt.ForegroundRole:
                if self.isAnalysisDir:
                    realRow = self._data.index[index.row()]
                    columnName = self._data.columns[index.column()]
                    if columnName == "L":
                        if self._data.isLoaded(realRow):
                            return QtCore.QVariant(QtGui.QColor("#4444EE"))
                    elif columnName == "A":
                        if self._data.analysisIsDirty(realRow):
                            # has been analyzed but not saved
                            return QtCore.QVariant(QtGui.QColor("#994444"))  # red
                        elif self._data.isAnalyzed(realRow):
                            return QtCore.QVariant(QtGui.QColor("#449944"))  # green
                    elif columnName == "S":
                        if self._data.isSaved(realRow):
                            return QtCore.QVariant(
                                QtGui.QColor("#999944")
                            )  # mustard yellow
                return QtCore.QVariant()
            elif role == QtCore.Qt.BackgroundRole:
                # set colors of alternating background
                # if index.row() % 2 == 0:
                #     return QtCore.QVariant(QtGui.QColor('#444444'))
                # else:
                #     return QtCore.QVariant(QtGui.QColor('#666666'))
                pass

        #
        return QtCore.QVariant()

    # def update(self, dataIn):
    #     print('  pandasModel.update() dataIn:', dataIn)

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        """
        Respond to user/keyboard edits.

        True if value is changed. Calls layoutChanged after update.
        Returns:
            False if value is not different from original value.
        """
        if index.isValid():
            if role == QtCore.Qt.EditRole:
                rowIdx = index.row()
                columnIdx = index.column()

                # use to check isEditable
                columnName = self._data.columns[columnIdx]
                if self.isAnalysisDir:
                    isEditable = self._data.columnIsEditable(columnName)
                    if not isEditable:
                        return False

                # in general, DO NOT USE iLoc, use loc as it is absolute (i,j)
                columnName = self._data.columns[index.column()]
                realRow = self._data.index[index.row()]
                v = self._data.loc[realRow, columnName]
                # logger.info(f'Existing value for column "{columnName}" is v: "{v}" {type(v)}')
                # logger.info(f'  proposed value:"{value}" {type(value)}')
                if isinstance(v, np.float64):
                    try:
                        if value == "":
                            value = np.nan
                        else:
                            value = float(value)
                    except ValueError as e:
                        logger.info("  No action -->> please enter a number")
                        # self.signalUpdateStatus.emit('Please enter a number')
                        return False

                # set
                # logger.info(f'  New value for column "{columnName}" is "{value}" {type(value)}')
                self._data.loc[realRow, columnName] = value
                # self._data.iloc[rowIdx, columnIdx] = value

                # emit change
                emitRowDict = self.myGetRowDict(realRow)
                self.signalMyDataChanged.emit(columnName, value, emitRowDict)

                self.isDirty = True
                return True
            elif role == QtCore.Qt.CheckStateRole:
                rowIdx = index.row()
                columnIdx = index.column()
                columnName = self._data.columns[index.column()]
                realRow = self._data.index[index.row()]
                logger.info(f"CheckStateRole column:{columnName} value:{value}")
                if columnName == "I":
                    self._data.loc[realRow, columnName] = value == 2
                    self.dataChanged.emit(index, index)
                    return QtCore.Qt.Checked

        #
        return QtCore.QVariant()

    def flags(self, index):
        if not index.isValid():
            logger.warning("index is not valid")

        rowIdx = index.row()
        columnIdx = index.column()

        # use to check isEditable
        try:
            columnName = self._data.columns[columnIdx]
        except IndexError as e:
            logger.warning(
                f"IndexError for columnIdx:{columnIdx} len:{len(self._data.columns)}"
            )
            print("self._data.columns:", self._data.columns)
            raise

        theRet = QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable
        isEditable = False
        isCheckbox = False
        if self.isAnalysisDir:
            # columnsDict is a big dict, one key for each column, in analysisDir.sanpyColumns
            isEditable = self._data.columnIsEditable(columnName)
            isCheckbox = self._data.columnIsCheckBox(columnName)
        if isEditable:
            theRet |= QtCore.Qt.ItemIsEditable
        if isCheckbox:
            # logger.info(f'isCheckbox {columnIdx}')
            theRet |= QtCore.Qt.ItemIsUserCheckable
        #
        return theRet

        # flags |= QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsUserCheckable | Qt.ItemIsEnabled

    def headerData(self, col, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                try:
                    return self._data.columns[col]
                except IndexError as e:
                    logger.warning(
                        f"IndexError for col:{col} len:{len(self._data.columns)}, shape:{self._data.shape}"
                    )
                    # raise
            elif orientation == QtCore.Qt.Vertical:
                # this is to show pandas 'index' column
                return col

        return QtCore.QVariant()

    def sort(self, Ncol, order):
        # logger.info(f'Ncol:{Ncol} order:{order}')
        self.layoutAboutToBeChanged.emit()
        if self.isAnalysisDir:
            self._data.sort_values(Ncol, order)
        else:
            self._data = self._data.sort_values(
                self._data.columns[Ncol], ascending=not order
            )
        self.layoutChanged.emit()

    def myCopyTable(self):
        """
        Copy model data to clipboard.
        """
        dfCopy = self._data.copy()
        dfCopy.to_clipboard(sep="\t", index=False)

    def myGetValue(self, rowIdx, colStr):
        val = None
        if colStr not in self._data.columns:  #  columns is a list
            logger.error(f'Got bad column name: "{colStr}"')
        elif len(self._data) - 1 < rowIdx:
            logger.error(f"Got bad row:{rowIdx} from possible {len(self._data)}")
        else:
            val = self._data.loc[rowIdx, colStr]
        return val

    def myGetRowDict(self, rowIdx):
        """
        return a dict with selected row as dict (includes detection parameters)
        """
        theRet = {}
        for column in self._data.columns:
            theRet[column] = self.myGetValue(rowIdx, column)
        return theRet

    def myGetColumnList(self, col):
        # return all values in column as a list
        colList = self._data[col].tolist()
        return colList

    def myAppendRow(self, rowDict=None):
        # append one empty row
        newRowIdx = len(self._data)
        self.beginInsertRows(QtCore.QModelIndex(), newRowIdx, newRowIdx)

        if self.isAnalysisDir:
            # if using analysis dir, azll actions are in-place
            self._data.appendRow()
        else:
            df = self._data
            df = df.append(pd.Series(), ignore_index=True)
            df = df.reset_index(drop=True)
            self._data = df

        self.endInsertRows()

    def old_myDeleteRow(self, rowIdx):
        msg = QtWidgets.QMessageBox()
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        msg.setDefaultButton(QtWidgets.QMessageBox.Ok)
        msg.setIcon(QtWidgets.QMessageBox.Warning)
        msg.setText(f"Are you sure you want to delete row {rowIdx}?")
        msg.setWindowTitle("Delete Row")
        returnValue = msg.exec_()
        if returnValue == QtWidgets.QMessageBox.Ok:
            #
            #
            rowIdx = self._data.index[rowIdx]  # assume rows are sorted
            self.beginRemoveRows(QtCore.QModelIndex(), rowIdx, rowIdx)
            #
            if self.isAnalysisDir:
                # if using analysis dir, azll actions are in-place
                self._data.deleteRow(rowIdx)
            else:
                df = self._data
                df = df.drop([rowIdx])
                df = df.reset_index(drop=True)
                self._data = df
            #
            self.endRemoveRows()

    def myUnloadRow(self, rowIdx):
        """Unload raw data by setting column ['_ba'] = None"""
        rowIdx = self._data.index[rowIdx]  # assume rows are sorted
        # self._data.loc[rowIdx, '_ba'] = None
        if self.isAnalysisDir:
            # if using analysis dir, azll actions are in-place
            self._data.unloadRow(rowIdx)

            # we changed the model, we need to emit dataChanged
            indexStart = self.createIndex(rowIdx, 0)
            indexStop = self.createIndex(rowIdx + 1, 0)
            self.dataChanged.emit(indexStart, indexStop)

    def myRemoveFromDatabase(self, rowIdx):
        """Remove bAnalysis from h5 database"""
        rowIdx = self._data.index[rowIdx]  # assume rows are sorted
        # self.beginInsertRows(QtCore.QModelIndex(), rowIdx, rowIdx)
        if self.isAnalysisDir:
            # if using analysis dir, azll actions are in-place
            # self.beginResetModel()
            self._data.removeRowFromDatabase(rowIdx)
            # self.endResetModel()

            # we changed the model, we need to emit dataChanged
            indexStart = self.createIndex(rowIdx, 0)
            indexStop = self.createIndex(rowIdx + 1, 0)
            self.dataChanged.emit(indexStart, indexStop)
        # self.endInsertRows()

    def myDuplicateRow(self, rowIdx):
        rowIdx = self._data.index[rowIdx]  # assume rows are sorted
        self.beginInsertRows(QtCore.QModelIndex(), rowIdx + 1, rowIdx + 1)
        #
        # duplicate rowIdx

        if self.isAnalysisDir:
            df = self._data._df  # either Dataframe or analysisDir
            self._data.duplicateRow(rowIdx)
        else:
            # make copy
            rowDict = self.myGetRowDict(rowIdx)
            newIdx = rowIdx + 0.5
            dfRow = pd.DataFrame(rowDict, index=[newIdx])

            df = self._data  # either Dataframe or analysisDir

            # append dfRow to the end
            df = df.append(dfRow, ignore_index=True)

            # sort by file name
            df = df.sort_values(
                by=["File"], axis="index", ascending=True, inplace=False
            )
            df = df.reset_index(drop=True)
            #
            self._data = df
        #
        self.endInsertRows()

    def mySetRow(self, rowIdx, rowDict):
        """Only set keys already in self._data.columns"""
        rowIdx = self._data.index[rowIdx]  # assume rows are sorted
        # rowSeries = pd.Series(rowDict)
        # self._data.loc[rowIdx] = rowSeries
        # self._data = self._data.reset_index(drop=True)
        for k, v in rowDict.items():
            if k in self._data.columns:
                self._data.at[rowIdx, k] = v

    def mySaveDb(self, path):
        # print('pandasModel.mySaveDb() path:', path)
        # logger.info(f'Saving csv {path}')
        self._data.to_csv(path, index=False)
        self.isDirty = False

    """
    def saveHdf(self):
        if self.isAnalysisDir:
            self._data.saveHdf()
    """

    def mySave(self):
        if self.isAnalysisDir:
            self._data.save()

    def mySyncDfWithPath(self):
        if self.isAnalysisDir:
            self.beginResetModel()
            self._data.syncDfWithPath()
            self.endResetModel()

    def myUpdateLoadedAnalyzed(self, ba, rowIdx):
        if self.isAnalysisDir:
            rowIdx = self._data.index[rowIdx]  # assume rows are sorted
            self._data._updateLoadedAnalyzed(rowIdx)

            # we changed the model, we need to emit dataChanged
            indexStart = self.createIndex(rowIdx, 0)
            indexStop = self.createIndex(rowIdx + 1, 0)
            self.dataChanged.emit(indexStart, indexStop)

    def _old_mySetDetectionParams(self, rowIdx, cellType):
        """Set predefined detection paramers.

        For example: SA Node, Neuron, Subthreshold, etc
        """
        logger.info(f"rowIdx:{rowIdx} cellType:{cellType}")
        rowDict = self.myGetRowDict(rowIdx)
        # print('  [1] rowDict:', rowDict)
        # get defaults
        dDict = sanpy.bAnalysis.getDefaultDetection(cellType=cellType)
        for k, v in dDict.items():
            if k in rowDict.keys():
                rowDict[k] = v
        # print('  [2] rowDict:', rowDict)
        self.mySetRow(rowIdx, rowDict)
        # print(self._data._df['_ba'])

        # signal data change
        indexStart = self.createIndex(rowIdx, 0)
        indexStop = self.createIndex(rowIdx + 1, 0)
        self.dataChanged.emit(indexStart, indexStop)

        # detection widget needs to update (dvdtThreshold, mVThreshold, ...)
        # emitRowDict = self.myGetRowDict(rowIdx)
        self.signalMyDataChanged.emit(None, None, rowDict)

    '''
    def mySetColumns(self, columnsDict):
        """
        When used as a file table, set with: sanpy.analysisDir.sanpyColumns
        """
        self.sanpyColumns = columnsDict
    '''


# see: https://stackoverflow.com/questions/17748546/pyqt-column-of-checkboxes-in-a-qtableview
class old_myCheckBoxDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QCheckBox cell of the column to which it's applied.
    """

    def __init__(self, parent):
        QtWidgets.QItemDelegate.__init__(self, parent)

    def createEditor(self, parent, option, index):
        """
        Important, otherwise an editor is created if the user clicks in this cell.
        """
        return None

    def paint(self, painter, option, index):
        """
        Paint a checkbox without the label.
        """
        self.drawCheck(
            painter,
            option,
            option.rect,
            QtCore.Qt.Unchecked if int(index.data()) == 0 else QtCore.Qt.Checked,
        )

    def editorEvent(self, event, model, option, index):
        """
        Change the data in the model and the state of the checkbox
        if the user presses the left mousebutton and this cell is editable. Otherwise do nothing.
        """
        if not int(index.flags() & QtCore.Qt.ItemIsEditable) > 0:
            return False

        if (
            event.type() == QtCore.QEvent.MouseButtonRelease
            and event.button() == QtCore.Qt.LeftButton
        ):
            # Change the checkbox-state
            self.setModelData(None, model, index)
            return True

        return False

    def setModelData(self, editor, model, index):
        """
        The user wanted to change the old state in the opposite.
        """
        print("myCheckBoxDelegate.setModelData()")
        model.setData(index, 1 if int(index.data()) == 0 else 0, QtCore.Qt.EditRole)



import glob
import importlib

# import inspect
import os
import traceback  # to print call stack on exception
from typing import List, Union

import sanpy

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


# TODO: put this in _util.py
def _module_from_file(module_name, file_path):
    """Load a module from a file.

    Parameters
    ----------
    module_name : str
        Name of the module.
    file_path : str
        Full path to the file

    Returns
    -------
    module
    """
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def _getObjectList(verbose=False) -> List[dict]:
    """Return a list of classes defined in sanpy.userAnalysis.

    Each of these is an object we can (i) construct or (ii) interrogate statis class members

    Returns
    -------
    list of dict
    """

    verbose = False

    if verbose:
        logger.info("")

    #
    # user plugins from files in folder <user>/SanPy/analysis
    userAnalysisFolder = sanpy._util._getUserAnalysisFolder()
    files = glob.glob(os.path.join(userAnalysisFolder, "*.py"))

    pluginDict = {}
    loadedModuleList = []

    for file in files:
        if file.endswith("__init__.py"):
            continue

        moduleName = os.path.split(file)[1]
        moduleName = os.path.splitext(moduleName)[0]
        fullModuleName = "sanpy.user_analysis." + moduleName  # + "." + moduleName

        loadedModule = _module_from_file(fullModuleName, file)

        if verbose:
            logger.info("")
            logger.info(f"    file: {file}")
            logger.info(f"    fullModuleName: {fullModuleName}")
            logger.info(f"    moduleName: {moduleName}")
            logger.info(f"    loadedModule: {loadedModule}")

        # class based user analysis
        oneConstructor = None
        try:
            oneConstructor = getattr(
                loadedModule, moduleName
            )  # moduleName is derived from file name (must match)
            if verbose:
                logger.info(f"    oneConstructor: {oneConstructor}")
                logger.info(f"    type(oneConstructor): {type(oneConstructor)}")
        except AttributeError as e:
            logger.error(
                f'Make sure filename and class name are the same,  file name is "{moduleName}"'
            )

        # instantiate the object and it will create a dictionary of new stats
        _tmpObj = oneConstructor(ba=None)
        _statStatDict = _tmpObj._getUserStatDict()

        # humanName = oneConstructor.myHumanName
        pluginDict = {
            "pluginClass": moduleName,
            "type": "user",
            "module": fullModuleName,
            "path": file,
            "constructor": oneConstructor,
            "staticStatDict": _statStatDict,
        }

        if verbose:
            logger.info(f'loading user analysis from file: "{file}"')

        loadedModuleList.append(pluginDict)

    # print out the entire list
    # logger.info('')
    # for loadedModuleDict in loadedModuleList:
    #     for k,v in loadedModuleDict.items():
    #         logger.info(f'    {k} : {v}')
    #
    return loadedModuleList  # list of dict


def findUserAnalysisStats() -> List[dict]:
    """Get the stat names of all user defined analysis."""
    userStatList: List[dict] = []
    objList = _getObjectList()  # list of dict
    for obj in objList:
        # sanpy._util.pprint(obj)
        # print('')

        # instantiate the object
        userObj = obj["constructor"](ba=None)

        userObjStatDict = userObj._getUserStatDict()

        for k, v in userObjStatDict.items():
            oneUserStatDict = {k: v}
            userStatList.append(oneUserStatDict)

    return userStatList


def runAllUserAnalysis(ba, verbose=False):
    """Run all user defined analysis.

    Notes
    -----
    Called at end of sanpy.bAnalysis.detect()
    """

    # step through each
    objList = _getObjectList()  # list of dict

    if verbose:
        logger.info(f"objList: {objList}")
    for obj in objList:
        # instantiate and call run (will add values for stats
        # was this
        # userObj = obj(ba)
        # userObj.run()

        try:
            # instantiate a user object
            userObj = obj["constructor"](ba)

            # run the analysis
            userObj.run()  # run the analysis and append to actual ba object
        except Exception as e:
            logger.error(f"Exception in running user defined analysis: {e}")
            logger.error(traceback.format_exc())


class baseUserAnalysis:
    """Create a userAnalysis object after bAnalysis has been analyzed with the core analysis results."""

    def __init__(self, ba: "sanpy.bAnalysis"):
        self._myAnalysis: sanpy.bAnalysis = ba

        self._userStatDict: dict = {}
        # add to this with addUserStat()

        self.defineUserStats()

    def _getUserStatDict(self):
        """Get dict of user defined stats, one key per stat."""
        return self._userStatDict

    def defineUserStats(self):
        """Derived classes add each stat with addUserStat().

        See Also
        --------
        addUserStat
        """
        pass

    def addUserStat(self, humanName: str, internalName: str):
        """Add a user stat. Derived classes do this in defineUserStats().

        Parameters
        ----------
        humanName : str
            Human readable name for the stat, like 'Threshold Potential (mV)'
        internalName : str
            Name to use for the variable name of the stat.
            Should not contain special characters like space or '-'
            Can contain '_'

        Notes
        -----
        userStatDict = {
            'User Time To Peak (ms)' : {
                'name': 'user_timeToPeak_ms',
                'units': 'ms',
                'yStat': 'user_timeToPeak_ms',
                'yStatUnits': 'ms',
                'xStat': 'thresholdPnt',
                'xStatUnits': 'Points'
                }
        }
        """
        if humanName in self._userStatDict.keys():
            logger.error(f'User stat with human name "{humanName}" already exists')
            return
        statDict = {
            "name": internalName,
            "units": None,
            "yStat": None,
            "yStatUnits": None,
            "xStat": None,
            "xStatUnits": None,
        }
        self._userStatDict[humanName] = statDict

    @property
    def ba(self):
        """Get the underlying [sanpy.bAnalysis][sanpy.bAnalysis] object"""
        return self._myAnalysis

    def getSweepX(self):
        """Get the x-axis of a recording."""
        return self.ba.fileLoader.sweepX

    def getSweepY(self):
        """Get the y-axis of a recording."""
        return self.ba.fileLoader.sweepY

    def getSweepC(self):
        """Get the DAC axis of a recording."""
        return self.ba.fileLoader.sweepC

    def getFilteredVm(self):
        return self.ba.fileLoader.sweepY_filtered

    def setSpikeValue(self, spikeIdx, theKey, theVal):
        """Set the value of a spike key.

        Parameters
        ----------
        spikeIdx : int
            The spike index , 0 based.
        theKey : str
            Name of the user defined internal name.
        theVal :
            The value for the key, can be almost any type like
            (float, int, bool, dict, list)

        Raises
        ------
        KeyError
            If theKey is not a key in analysis results.
        IndexError
            If spikeIdx is beyond number of spikes -1.
        """
        try:
            self.ba.spikeDict[spikeIdx][theKey] = theVal
        except KeyError as e:
            logger.error(f'User internal stat does not exist "{theKey}"')
        except IndexError as e:
            logger.error(
                f"spikeIdx {spikeIdx} is out of range, max value is {self.ba.numSpikes}"
            )

    def getSpikeValue(self, spikeIdx, theKey):
        """Get a single spike analysis result from key.

        Parameters
        ----------
        spikeIdx : int
            The spike index, 0 based.
        theKey : str
            Name of the analysis result defined internal name.

        Raises
        ------
        KeyError
            If theKey is not a key in analysis results.
        IndexError
            If spikeIdx is beyond number of spikes -1.
        """
        try:
            theRet = self.ba.spikeDict[spikeIdx][theKey]
            return theRet
        except KeyError as e:
            logger.error(f'User internal stat does not exist "{theKey}"')
        except IndexError as e:
            logger.error(
                f"spikeIdx {spikeIdx} is out of range, max value is {self.ba.numSpikes}"
            )

    def run(self):
        """Run user analysis. Calculate values for each new user stat."""


if __name__ == "__main__":
    # test1()
    # _getObjectList()
    findUserAnalysisStats()



import numpy as np
import pandas as pd

from sanpy.fileloaders.fileLoader_base import fileLoader_base, recordingModes

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class fileLoader_csv(fileLoader_base):
    loadFileType = "csv"

    # @property
    # def loadFileType(self):
    #     return 'csv'

    def loadFile(self):
        """Load file and call setLoadedData().

        Use self.filepath for the file

        Column 0 is seconds
        Column 1 is recording in ['mV', 'pA']
        """
        # load the csv file
        try:
            _df = pd.read_csv(self.filepath)
        except (pd.errors.EmptyDataError):
            self._loadError = True
            return
        
        _columns = _df.columns
        _numColumns = len(_columns)
        if _numColumns < 2:
            logger.error(f"expecting 2 or more columns but got {_numColumns}")
            self._loadError = True
            return

        _numSamples = len(_df)
        _numSweeps = _numColumns - 1  # assuming no DAC columns

        # TODO: check that the columns we will read are float (not str or object)
        #
        # make sweepX (assuming all sweeps have the same sweepX)
        sweepX = np.ndarray((_numSamples, 1))
        try:
            sweepX[:, 0] = _df[_columns[0]].to_numpy()
        except ValueError as e:
            # raise ValueError
            logger.error(f"ValueError assigning the first column")
            self._loadError = True
            return

        # make sweepY to hold (samples, sweeps) values, the values in each sweep are different
        sweepY = np.ndarray((_numSamples, _numSweeps))
        for sweep in range(_numSweeps):
            colIdx = sweep + 1
            sweepY[:, sweep] = _df[_columns[colIdx]].to_numpy()

        # sweepC = None  # not defined

        # determine recording mode from name of column 1
        _tmpColName = _columns[1].replace("_0", "")
        if _tmpColName in ["pA", "nA"]:
            _recordingMode = recordingModes.vclamp
        elif _tmpColName in ["mV", "mv"]:
            _recordingMode = recordingModes.iclamp
        else:
            logger.warning(
                f"did not infer recording mode from column name {_tmpColName}"
            )
            _recordingMode = recordingModes.unknown

        xLabel = _columns[0]
        yLabel = _tmpColName  # _columns[1]

        self.setLoadedData(
            sweepX=sweepX,
            sweepY=sweepY,
            # sweepC = sweepC,
            xLabel=xLabel,
            yLabel=yLabel,
        )



import math

from typing import Union, Dict, List, Tuple, Optional

import numpy as np
import pandas as pd

from PyQt5 import QtCore, QtWidgets, QtGui

from matplotlib.backends import backend_qt5agg
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.widgets import RectangleSelector  # To click+drag rectangular selection
import matplotlib.markers as mmarkers  # To define different markers for scatter

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin

def getPlotMarkersAndColors(ba : sanpy.bAnalysis,
                            spikeList : List[int],
                            hue = '') -> dict:
    """Given a list of spikes, get plotting color and symbol.
    
    Parse bAnalysis for 'condition' and usertype'

    Parameters
    ----------
    ba : sanpy.bAnalysis
    spikeList List[int]
    hue : str
        In ("", "Time", "Sweep")

    TODO: Add one more return for pyqtgraph markers in ('o', 'star', 't', 't3')
    """
    cMap = mpl.pyplot.cm.coolwarm.copy()
    cMap.set_under("white")  # only works for dark theme
    # do not specify 'c' argument, we set colors using set_facecolor, set_color

    cMap = None
    colorMapArray = None
    faceColors = None
    pathList = None
    markerList_pg = None

    if hue == "Time":
        colorMapArray = np.array(range(len(spikeList)))
        # self.lines.set_array(colorMapArray)  # set_array is for a color map
        # self.lines.set_cmap(cmap)  # mpl.pyplot.cm.coolwarm
        # self.lines.set_color(faceColors)

    elif hue == "Sweep":
        # color sweeps
        _sweeps = ba.getSpikeStat(spikeList, 'sweep')
        colorMapArray = np.array(range(len(_sweeps)))
        # self.lines.set_array(colorMapArray)  # set_array is for a color map
        # self.lines.set_cmap(self.cmap)  # mpl.pyplot.cm.coolwarm
        # self.lines.set_color(faceColors)

    else:
        # tmpColor = np.array(range(len(xData)))
        # assuming self.cmap.set_under("white")

        # from matplotlib.colors import ListedColormap
        #_color_dict = {1: "blue", 2: "red", 13: "orange", 7: "green"}
        _color_dict = {
            "good": 'c',  # cyan
            "bad": 'r',  # red
            #'userType1':(0,0,1,1), # blue
            #'userType2':(0,1,1,1), # cyan
            #'userType3':(1,0,1,1), # magenta
        }
        _marker_dict = {
            "noUserType": mmarkers.MarkerStyle("o"),  # circle
            "userType1": mmarkers.MarkerStyle("*"),  # star
            "userType2": mmarkers.MarkerStyle("v"),  # triangle_down
            "userType3": mmarkers.MarkerStyle("<"),  # triangle_left
        }

        _marker_dict_pg = {
            "noUserType": 'o',
            "userType1": 'star',
            "userType2": 't',
            "userType3": 't3',
        }

        # no need for a cmap ???
        # cm = ListedColormap([_color_dict[x] for x in _color_dict.keys()])
        # self.lines.set_cmap(cm)

        goodSpikes = ba.getSpikeStat(spikeList, "include")
        #print('goodSpikes:', goodSpikes)
        userTypeList = ba.getSpikeStat(spikeList, "userType")

        # logger.warning("   debug set spike stat, user types need to be int")
        # print('userTypeList:', type(userTypeList))
        # print(userTypeList)

        # user types will use symbols
        # tmpColors = [_color_dict['type'+num2str(x)] if x>0 else tmpColors[idx] for idx,x in enumerate(userTypeList)]
        # bad needs to trump user type !!!
        # faceColors is used to set_facecolor() and set_color
        if goodSpikes is not None:
            faceColors = [
                _color_dict["good"] if x else _color_dict["bad"]
                for idx, x in enumerate(goodSpikes)  # x is (good, bad), (1,0)
            ]
        #faceColors = np.array(faceColors)
        # print('tmpColors', type(tmpColors), tmpColors.shape, tmpColors)

        # self.lines.set_array(None)  # used to map [0,1] to color map
        # self.lines.set_facecolor(faceColors)
        # self.lines.set_color(faceColors)  # sets the outline

        # set user type 2 to 'star'
        # see: https://stackoverflow.com/questions/52303660/iterating-markers-in-plots/52303895#52303895
        # import matplotlib.markers as mmarkers
        markerList = []
        markerList_pg = []
        if userTypeList is not None:
            for x in userTypeList:
                markerList.append(
                    _marker_dict["userType" + str(x)] if x > 0 else _marker_dict['noUserType']
                )
                markerList_pg.append(
                    _marker_dict_pg["userType" + str(x)] if x > 0 else _marker_dict_pg['noUserType']
                )
            
        # tODO: roll this into above for loop
        pathList = []
        for marker in markerList:
            path = marker.get_path().transformed(marker.get_transform())
            pathList.append(path)
        #self.lines.set_paths(pathList)

    retDict = {
        'cMap': cMap,
        'colorMapArray': colorMapArray,
        'faceColors': faceColors,
        'pathList': pathList,
        'markerList_pg': markerList_pg,
    }

    # logger.info(f'spikeList: {spikeList}')
    # for k,v in retDict.items():
    #     print('    ', k, ':', v)

    return retDict

class plotScatter(sanpyPlugin):
    """Plot x/y statiistics as a scatter.

    Get stat names and variables from sanpy.bAnalysisUtil.getStatList()
    """

    myHumanName = "Plot Scatter"

    def __init__(self, **kwargs):
        """
        Args:
            ba (bAnalysis): Not required
        """
        super().__init__(**kwargs)

        self._hueList = ["None", "Time", "Sweep"]
        self._hue = "None"  # from ["None", "Time", "Sweep"]
        
        # april 2023, deactivated this, need to debug
        self.plotChasePlot = False  # i vs i-1
        
        # self.plotIsBad= True
        self.plotHistograms = True

        # keep track of what we are plotting.
        # use this in replot() and copy to clipboard.
        self.xStatName = None
        self.yStatName = None
        self.xStatHumanName = None
        self.yStatHumanName = None

        self.xData = []
        self.yData = []
        
        self._markerSize = 20

        # when we plot, we plot a subset of spikes
        # this list tells us the spikes we are plotting and [ind_1] gives us the real spike number
        self._plotSpikeNumber = []
        
        # self.xAxisSpikeNumber = []  # We need to keep track of this for 'Chase Plot'

        # main layout
        hLayout = QtWidgets.QHBoxLayout()
        hSplitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        hLayout.addWidget(hSplitter)

        # coontrols and 2x stat list
        vLayout = QtWidgets.QVBoxLayout()

        # controls
        controlWidget = QtWidgets.QWidget()

        hLayout2 = QtWidgets.QHBoxLayout()

        #
        # self.b2 = QtWidgets.QCheckBox("Chase Plot")
        # self.b2.setEnabled(False)  # april 15, 2023, disbale for now (need to debug)
        # self.b2.setChecked(self.plotChasePlot)
        # self.b2.stateChanged.connect(lambda: self.btnstate(self.b2))
        # hLayout2.addWidget(self.b2)

        #
        # self.colorTime = QtWidgets.QCheckBox("Color Time")
        # self.colorTime.setChecked(self.plotColorTime)
        # self.colorTime.stateChanged.connect(lambda:self.btnstate(self.colorTime))
        # hLayout2.addWidget(self.colorTime)

        aLabel = QtWidgets.QLabel("Hue")
        hLayout2.addWidget(aLabel)
        aComboBox = QtWidgets.QComboBox()
        aComboBox.addItems(self._hueList)  # from none, time, sweep
        aComboBox.currentTextChanged.connect(self._on_select_hue)
        hLayout2.addWidget(aComboBox)

        #
        """
        self.colorType = QtWidgets.QCheckBox("Color Type")
        self.colorType.setChecked(self.plotColorType)
        self.colorType.stateChanged.connect(lambda:self.btnstate(self.colorType))
        hLayout2.addWidget(self.colorType)
        """

        #
        """
        self.showBad = QtWidgets.QCheckBox("Show Bad")
        self.showBad.setChecked(self.plotIsBad)
        self.showBad.stateChanged.connect(lambda:self.btnstate(self.showBad))
        hLayout2.addWidget(self.showBad)
        """

        #
        self.histogramCheckbox = QtWidgets.QCheckBox("Histograms")
        self.histogramCheckbox.setChecked(self.plotHistograms)
        self.histogramCheckbox.stateChanged.connect(
            lambda: self.btnstate(self.histogramCheckbox)
        )
        hLayout2.addWidget(self.histogramCheckbox)

        vLayout.addLayout(hLayout2)

        # second row of controls
        hLayout2_2 = QtWidgets.QHBoxLayout()

        aName = "Spike: None"
        self.spikeNumberLabel = QtWidgets.QLabel(aName)
        hLayout2_2.addWidget(self.spikeNumberLabel)

        vLayout.addLayout(hLayout2_2)

        # x and y stat lists
        hLayout3 = QtWidgets.QHBoxLayout()
        self.xPlotWidget = myStatListWidget(self, headerStr="X Stat")
        self.xPlotWidget.myTableWidget.selectRow(0)
        self.yPlotWidget = myStatListWidget(self, headerStr="Y Stat")
        self.yPlotWidget.myTableWidget.selectRow(7)
        hLayout3.addWidget(self.xPlotWidget)
        hLayout3.addWidget(self.yPlotWidget)
        vLayout.addLayout(hLayout3)

        controlWidget.setLayout(vLayout)

        #
        # create a mpl plot (self._static_ax, self.static_canvas)
        # self.mplWindow2()
        # plt.style.use('dark_background')
        if self.darkTheme:
            plt.style.use("dark_background")
        else:
            plt.rcParams.update(plt.rcParamsDefault)

        # this is dangerous, collides with self.mplWindow()
        self.fig = mpl.figure.Figure()
        self.static_canvas = backend_qt5agg.FigureCanvas(self.fig)
        self.static_canvas.setFocusPolicy(
            QtCore.Qt.ClickFocus
        )  # this is really tricky and annoying
        self.static_canvas.setFocus()
        # self.axs[idx] = self.static_canvas.figure.add_subplot(numRow,1,plotNum)

        self._switchScatter()
        """
        # gridspec for scatter + hist
        self.gs = self.fig.add_gridspec(2, 2,  width_ratios=(7, 2), height_ratios=(2, 7),
                                left=0.1, right=0.9, bottom=0.1, top=0.9,
                                wspace=0.05, hspace=0.05)

        self.axScatter = self.static_canvas.figure.add_subplot(self.gs[1, 0])
        self.axHistX = self.static_canvas.figure.add_subplot(self.gs[0, 0], sharex=self.axScatter)
        self.axHistY = self.static_canvas.figure.add_subplot(self.gs[1, 1], sharey=self.axScatter)


        # make initial empty scatter plot
        #self.lines, = self._static_ax.plot([], [], 'ow', picker=5)
        self.cmap = mpl.pyplot.cm.coolwarm
        self.cmap.set_under("white") # only works for dark theme
        #self.lines = self._static_ax.scatter([], [], c=[], cmap=self.cmap, picker=5)
        self.lines = self.axScatter.scatter([], [], c=[], cmap=self.cmap, picker=5)

        # make initial empty spike selection plot
        #self.spikeSel, = self._static_ax.plot([], [], 'oy')
        self.spikeSel, = self.axScatter.plot([], [], 'oy')

        # despine top/right
        self.axScatter.spines['right'].set_visible(False)
        self.axScatter.spines['top'].set_visible(False)
        self.axHistX.spines['right'].set_visible(False)
        self.axHistX.spines['top'].set_visible(False)
        self.axHistY.spines['right'].set_visible(False)
        self.axHistY.spines['top'].set_visible(False)
        """

        # can do self.mplToolbar.hide()
        # matplotlib.backends.backend_qt5.NavigationToolbar2QT
        self.mplToolbar = mpl.backends.backend_qt5agg.NavigationToolbar2QT(
            self.static_canvas, self.static_canvas
        )

        # put toolbar and static_canvas in a V layout
        plotWidget = QtWidgets.QWidget()
        vLayoutPlot = QtWidgets.QVBoxLayout()
        vLayoutPlot.addWidget(self.static_canvas)
        vLayoutPlot.addWidget(self.mplToolbar)
        plotWidget.setLayout(vLayoutPlot)

        # finalize
        # hLayout.addLayout(vLayout)
        hSplitter.addWidget(controlWidget)
        # hSplitter.addWidget(self.static_canvas) # add mpl canvas
        hSplitter.addWidget(plotWidget)

        # set the layout of the main window
        # playing with dock
        # self.mainWidget.setLayout(hLayout)

        # mar 17 was this
        # self.setLayout(hLayout)
        self.getVBoxLayout().addLayout(hLayout)

        # self.mainWidget.setGeometry(100, 100, 1200, 600)

        self.replot()

    def keyPressEvent(self, event):
        _handled = False
        isMpl = isinstance(event, mpl.backend_bases.KeyEvent)
        if isMpl:
            text = event.key
            logger.info(f'mpl key: "{text}"')
            if text in ['=', '+']:
                # increase scatter points
                _handled = True
                self._markerSize += 10
                logger.info(f'marker size is now {self._markerSize}')
                self.replot()
            elif text in ['-']:
                _handled = True
                self._markerSize -= 10
                if self._markerSize<0:
                    self._markerSize = 0
                logger.info(f'marker size is now {self._markerSize}')
                self.replot()
        if not _handled:
            super().keyPressEvent(event)

    def _switchScatter(self):
        """Switch between single scatter plot and scatter + marginal histograms"""

        if self.plotHistograms:
            # gridspec for scatter + hist
            self.gs = self.fig.add_gridspec(
                2,
                2,
                width_ratios=(7, 2),
                height_ratios=(2, 7),
                left=0.1,
                right=0.9,
                bottom=0.1,
                top=0.9,
                wspace=0.05,
                hspace=0.05,
            )
        else:
            self.gs = self.fig.add_gridspec(
                1, 1, left=0.1, right=0.9, bottom=0.1, top=0.9, wspace=0.05, hspace=0.05
            )

        self.static_canvas.figure.clear()
        if self.plotHistograms:
            self.axScatter = self.static_canvas.figure.add_subplot(self.gs[1, 0])

            # x/y hist
            self.axHistX = self.static_canvas.figure.add_subplot(
                self.gs[0, 0], sharex=self.axScatter
            )
            self.axHistY = self.static_canvas.figure.add_subplot(
                self.gs[1, 1], sharey=self.axScatter
            )
            #
            self.axHistX.spines["right"].set_visible(False)
            self.axHistX.spines["top"].set_visible(False)
            self.axHistY.spines["right"].set_visible(False)
            self.axHistY.spines["top"].set_visible(False)

            # self.axHistX.tick_params(axis="x", labelbottom=False) # no labels
            # self.axHistX.tick_params(axis="y", labelleft=False) # no labels
        else:
            self.axScatter = self.static_canvas.figure.add_subplot(self.gs[0, 0])
            self.axHistX = None
            self.axHistY = None

        # we might have memory garbage collection issues
        # passing self so we can receive selectSpikesFromHighlighter()
        self.myHighlighter = Highlighter(self, self.axScatter, [], [], [])

        #
        # we will always have axScatter
        #
        # make initial empty scatter plot
        self.cmap = mpl.pyplot.cm.coolwarm.copy()
        self.cmap.set_under("white")  # only works for dark theme
        # do not specify 'c' argument, we set colors using set_facecolor, set_color

        self.lines = self.axScatter.scatter([], [], picker=5)

        # make initial empty spike selection plot

        # self.spikeSel, = self.axScatter.plot([], [], 'x', markerfacecolor='none', color='y', markersize=10)  # no picker for selection
        # TODO: standardize this with Highlighter scatter
        # was this
        # self.spikeListSel, = self.axScatter.plot([], [], 'o', markerfacecolor='none', color='y', markersize=10)  # no picker for selection
        #logger.info("hard coding spike selection markerSize=10")

        # markerSize = self._markerSize / 2  # 10
        # (self.spikeListSel,) = self.axScatter.plot(
        #     [], [], "o", markersize=markerSize, color="yellow", zorder=10
        # )

        # despine top/right
        self.axScatter.spines["right"].set_visible(False)
        self.axScatter.spines["top"].set_visible(False)

        # TODO: refactor and use mplToolBar()
        # self.cid = self.static_canvas.mpl_connect("pick_event", self._on_spike_pick_event2)

        self.fig.canvas.mpl_connect("key_press_event", self.keyPressEvent)

        #
        # self.replot()

    def _old__on_spike_pick_event2(self, event):
        """Respond to user left-mouse clicks in mpl plot.

        Args:
            event : matplotlib.backend_bases.PickEvent
                PickEvent with indices in ind[]

        Notes:
            need to clean up fft on pick, it is still in base plugin class
        """

        # ignore when not left mouse button
        if event.mouseevent.button != 1:
            return

        # no hits
        if len(event.ind) < 1:
            return

        _clickedPlotIdx = event.ind[0]

        # convert to what we are actually plotting
        try:
            #_realSpikeNumber = self._plotSpikeNumber.index(_clickedPlotIdx)
            # get real spike number from subset of plotted psikes
            _realSpikeNumber = self._plotSpikeNumber[_clickedPlotIdx]
        except (IndexError) as e:
            logger.warning(f'  xxx we are not plotting _clickedPlotIdx {_clickedPlotIdx}')

        doZoom = False
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ShiftModifier:
            doZoom = True

        #logger.info(f'  _clickedPlotIdx:{_clickedPlotIdx} _realSpikeNumber:{_realSpikeNumber}')
        logger.info(
            f"  got {len(event.ind)} candidates, first is spike:{_realSpikeNumber} doZoom:{doZoom}"
        )

        # propagate a signal to parent
        # TODO: use class SpikeSelectEvent()
        sDict = {
            "spikeList": [_realSpikeNumber],
            "doZoom": doZoom,
            "ba": self.ba,
        }
        logger.info(f'{self._myClassName()} -->> emit signalSelectSpikeList')
        self.signalSelectSpikeList.emit(sDict)

    def btnstate(self, b):
        state = b.isChecked()
        # if b.text() == "Respond To Analysis Changes":
        #    self.setRespondToAnalysisChange(state)
        if b.text() == "Chase Plot":
            self.plotChasePlot = state
            logger.info(f"plotChasePlot:{self.plotChasePlot}")
            self.replot()
        # elif b.text() == 'Color Time':
        #     self.plotColorTime = state
        #     logger.info(f'plotColorTime:{self.plotColorTime}')
        #     self.replot()

        # elif b.text() == 'Color Type':
        #    self.plotColorType = state
        #    logger.info(f'plotColorType:{self.plotColorType}')
        #    self.replot()
        # elif b.text() == 'Show Bad':
        #    self.plotIsBad = state
        #    logger.info(f'plotIsBad:{self.plotIsBad}')
        #    self.replot()
        elif b.text() == "Histograms":
            self.plotHistograms = state
            self._switchScatter()
            logger.info(f"plotHistograms:{self.plotHistograms}")
            self.replot()
        else:
            logger.warning(f'Did not respond to button "{b.text()}"')

    def _on_select_hue(self, hue: str):
        self._hue = hue
        self.replot()

    def setAxis(self):
        """Inherited, respond to user setting x-axis.

        Generate a list of spikes and select them.
        """

        logger.info('deactivated - april 30')

        # as of april 30, 2023, do nothin on set axis
        return
    
        startSec, stopSec = self.getStartStop()
        if startSec is None or stopSec is None:
            return

        # select spike in range
        thresholdSec = self.getStat("thresholdSec")
        _selectedSpikeList = [
            spikeIdx
            for spikeIdx, spikeSec in enumerate(thresholdSec)
            if (spikeSec > startSec and spikeSec < stopSec)
        ]
        self.setSelectedSpikes(_selectedSpikeList)
        self.selectSpikeList()

    def replot(self):
        """Replot when file or analysis changes.
        """

        logger.info(f"{self._myClassName()} sweepNumber:{self.sweepNumber} epochNumber:{self.epochNumber}")

        # get from stat lists
        xHumanStat, xStat = self.xPlotWidget.getCurrentStat()
        yHumanStat, yStat = self.yPlotWidget.getCurrentStat()

        # logger.info(f'x:"{xHumanStat}" y:"{yHumanStat}"')
        # logger.info(f'x:"{xStat}" y:"{yStat}"')

        if self.ba is None or xHumanStat is None or yHumanStat is None:
            xData = []
            yData = []
            self._plotSpikeNumber = []
        else:
            # use getFullList when we have a recording with
            # multiple sweeps and epochs
            xData = self.getStat(xStat, getFullList=False)
            yData = self.getStat(yStat, getFullList=False)
            self._plotSpikeNumber = self.getStat('spikeNumber')

        # logger.warning('debug plotting epochs')
        # print('xData:', xData)

        #
        # TODO: use ba.getStat() option for this.)
        xData = np.array(xData)
        yData = np.array(yData)

        #
        # return if we got no data, happens when there is no analysis
        if (
            xData is None
            or yData is None
            or np.isnan(xData).all()
            or np.isnan(yData).all()
        ):
            # We get here when there is no analysis
            # logger.warning(f'Did not find either xStat: "{xStat}" or yStat: "{yStat}"')
            self.lines.set_offsets([np.nan, np.nan])
            self.scatter_hist([], [], self.axHistX, self.axHistY)
            self.static_canvas.draw()
            return

        self.xStatName = xStat
        self.yStatName = yStat
        self.xStatHumanName = xHumanStat
        self.yStatHumanName = yHumanStat

        # need to mask color and marker
        if self.plotChasePlot:
            # keep track of x-axis spike number (for chase plot)
            xAxisSpikeNumber = self.getStat("spikeNumber")
            xAxisSpikeNumber = np.array(xAxisSpikeNumber)

            # on selection, ind will refer to y-axis spike
            xData = xData[1:-1]  # x is the reference spike for marking (bad, type)
            yData = yData[0:-2]
            #
            # xAxisSpikeNumber = xAxisSpikeNumber[1:-1]
            xAxisSpikeNumber = xAxisSpikeNumber[0:-2]

        self.xData = xData
        self.yData = yData
        # self.xAxisSpikeNumber = xAxisSpikeNumber

        # data
        data = np.stack([xData, yData], axis=1)
        self.lines.set_offsets(data)  # (N, 2)
        
        _sizes = [self._markerSize] * len(xData)
        self.lines.set_sizes(_sizes)

        # AttributeError: 'Line2D' object has no attribute 'set_sizes'
        # _sizes = [self._markerSize * 0.3] * len(xData)
        # self.spikeListSel.set_sizes()

        # 20230417
        plotSpikeNumberList = self.getStat('spikeNumber', getFullList=False)
        _tmpDict = getPlotMarkersAndColors(self.ba, plotSpikeNumberList, self._hue)

        cMap = _tmpDict['cMap']
        colorMapArray = _tmpDict['colorMapArray']
        faceColors = _tmpDict['faceColors']
        pathList = _tmpDict['pathList']

        self.lines.set_array(colorMapArray)  # set_array is for a color map
        self.lines.set_cmap(cMap)  # mpl.pyplot.cm.coolwarm
        self.lines.set_color(faceColors)
        self.lines.set_color(faceColors)  # sets the outline
        self.lines.set_paths(pathList)

        #
        # color
        # if self._hue == "Time":
        #     tmpColor = np.array(range(len(xData)))
        #     self.lines.set_array(tmpColor)  # set_array is for a color map
        #     self.lines.set_cmap(self.cmap)  # mpl.pyplot.cm.coolwarm
        #     self.lines.set_color(None)
        # elif self._hue == "Sweep":
        #     # color sweeps
        #     _sweeps = self.getStat("sweep")
        #     tmpColor = np.array(range(len(_sweeps)))
        #     self.lines.set_array(tmpColor)  # set_array is for a color map
        #     self.lines.set_cmap(self.cmap)  # mpl.pyplot.cm.coolwarm
        #     self.lines.set_color(None)

        # else:
        #     # tmpColor = np.array(range(len(xData)))
        #     # assuming self.cmap.set_under("white")

        #     # from matplotlib.colors import ListedColormap
        #     color_dict = {1: "blue", 2: "red", 13: "orange", 7: "green"}
        #     color_dict = {
        #         "good": (0, 1, 1, 1),  # cyan
        #         "bad": (1, 0, 0, 1),  # red
        #         #'userType1':(0,0,1,1), # blue
        #         #'userType2':(0,1,1,1), # cyan
        #         #'userType3':(1,0,1,1), # magenta
        #     }
        #     marker_dict = {
        #         "userType1": mmarkers.MarkerStyle("*"),  # star
        #         "userType2": mmarkers.MarkerStyle("v"),  # triangle_down
        #         "userType3": mmarkers.MarkerStyle("<"),  # triangle_left
        #     }

        #     # no need for a cmap ???
        #     # cm = ListedColormap([color_dict[x] for x in color_dict.keys()])
        #     # self.lines.set_cmap(cm)

        #     goodSpikes = self.getStat("include")
        #     userTypeList = self.getStat("userType")

        #     logger.warning("   debug set spike stat, user types need to be int")
        #     logger.warning("   we ARE NOT GETTING THE CORRECT SPIKE INDEX")
        #     # print(userTypeList)

        #     if self.plotChasePlot:
        #         # xData = xData[1:-1] # x is the reference spike for marking (bad, type)
        #         # goodSpikes = goodSpikes[1:-1]
        #         # userTypeList = userTypeList[1:-1]
        #         goodSpikes = goodSpikes[0:-2]
        #         userTypeList = userTypeList[0:-2]

        #     tmpColors = [color_dict["bad"]] * len(xData)  # start as all good

        #     # user types will use symbols
        #     # tmpColors = [color_dict['type'+num2str(x)] if x>0 else tmpColors[idx] for idx,x in enumerate(userTypeList)]
        #     # bad needs to trump user type !!!
        #     tmpColors = [
        #         color_dict["good"] if x else tmpColors[idx]
        #         for idx, x in enumerate(goodSpikes)
        #     ]
        #     tmpColors = np.array(tmpColors)
        #     # print('tmpColors', type(tmpColors), tmpColors.shape, tmpColors)

        #     self.lines.set_array(None)  # used to map [0,1] to color map
        #     self.lines.set_facecolor(tmpColors)
        #     self.lines.set_color(tmpColors)  # sets the outline

        #     # set user type 2 to 'star'
        #     # see: https://stackoverflow.com/questions/52303660/iterating-markers-in-plots/52303895#52303895
        #     # import matplotlib.markers as mmarkers
        #     myMarkerList = [
        #         marker_dict["userType" + str(x)] if x > 0 else mmarkers.MarkerStyle("o")
        #         for x in userTypeList
        #     ]
        #     myPathList = []
        #     for myMarker in myMarkerList:
        #         path = myMarker.get_path().transformed(myMarker.get_transform())
        #         myPathList.append(path)
        #     self.lines.set_paths(myPathList)

        #
        # update highlighter, needs coordinates of x/y to highlight
        self.myHighlighter.setData(xData, yData, self._plotSpikeNumber)

        # label axes
        xStatLabel = xHumanStat
        yStatLabel = yHumanStat
        if self.plotChasePlot:
            xStatLabel += " [i]"
            yStatLabel += " [i-1]"
        self.axScatter.set_xlabel(xStatLabel)
        self.axScatter.set_ylabel(yStatLabel)

        # don't cancel on replot
        """
        # cancel any selections
        self.spikeSel.set_data([], [])
        #self.spikeSel.set_offsets([], [])
        self.spikeListSel.set_data([], [])
        #self.spikeListSel.set_offsets([], [])
        """

        xMin = np.nanmin(xData)
        xMax = np.nanmax(xData)
        yMin = np.nanmin(yData)
        yMax = np.nanmax(yData)
        # expand by 5%
        xSpan = abs(xMax - xMin)
        percentSpan = xSpan * 0.05
        xMin -= percentSpan
        xMax += percentSpan
        #
        ySpan = abs(yMax - yMin)
        percentSpan = ySpan * 0.05
        yMin -= percentSpan
        yMax += percentSpan

        #logger.warning(f'refactor this, do not always set x/y lim')
        # UserWarning: Attempting to set identical left == right == 0.0275 results in singular transformations; automatically expanding.
        self.axScatter.set_xlim([xMin, xMax])
        self.axScatter.set_ylim([yMin, yMax])

        # self.scatter_hist(xData, yData, self.axScatter, self.axHistX, self.axHistY)
        self.scatter_hist(xData, yData, self.axHistX, self.axHistY)

        # redraw
        self.static_canvas.draw()
        # was this
        # self.repaint() # update the widget

    # def scatter_hist(self, x, y, ax, ax_histx, ax_histy):
    def scatter_hist(self, x, y, ax_histx, ax_histy):
        """
        plot a scatter with x/y histograms in margin

        Args:
            x (date):
            y (data):
            ax_histx (axes) Histogram Axes
            ax_histy (axes) Histogram Axes
        """

        xBins = "auto"
        yBins = "auto"

        xTmp = np.array(x)  # y[~np.isnan(y)]
        xTmp = xTmp[~np.isnan(xTmp)]
        xTmpBins = np.histogram_bin_edges(xTmp, "auto")
        xNumBins = len(xTmpBins)
        if xNumBins * 2 < len(x):
            xNumBins *= 2
        xBins = xNumBins

        yTmp = np.array(y)  # y[~np.isnan(y)]
        yTmp = yTmp[~np.isnan(yTmp)]
        yTmpBins = np.histogram_bin_edges(yTmp, "auto")
        yNumBins = len(yTmpBins)
        if yNumBins * 2 < len(y):
            yNumBins *= 2
        yBins = yNumBins

        # x
        if ax_histx is not None:
            ax_histx.clear()
            nHistX, binsHistX, patchesHistX = ax_histx.hist(
                x, bins=xBins, facecolor="silver", edgecolor="gray"
            )
            ax_histx.tick_params(axis="x", labelbottom=False)  # no labels
            # ax_histx.spines['right'].set_visible(False)
            # ax_histx.spines['top'].set_visible(False)
            # ax_histx.yaxis.set_ticks_position('left')
            # ax_histx.xaxis.set_ticks_position('bottom')
            # print('  binsHistX:', len(binsHistX))
        # y
        if ax_histy is not None:
            ax_histy.clear()
            nHistY, binsHistY, patchesHistY = ax_histy.hist(
                y,
                bins=yBins,
                orientation="horizontal",
                facecolor="silver",
                edgecolor="gray",
            )
            ax_histy.tick_params(axis="y", labelleft=False)
            # ax_histy.yaxis.set_ticks_position('left')
            # ax_histy.xaxis.set_ticks_position('bottom')
            # print('  binsHistY:', len(binsHistY))

    def selectSpikeList(self):
        """Use getSelectedSpikes() to select spikes.
        
        Notes
        -----
        Always have to use the list of spikes we are actually plotting
        self._plotSpikeNumber
        """
        spikeList = self.getSelectedSpikes()

        logger.info(f"{self._myClassName()} spikeList:{len(spikeList)} {self}")

        if self.xStatName is None or self.yStatName is None:
            return

        # TODO (error) when we are plotting just one epoch (like 2), we get index errors
        
        # if spikeList is not None:
        _plotSpikeIndexList = []
        if len(spikeList) > 0:
            for _realSpikeNumber in spikeList:
                # _realSpikeNumber is the actual spike number of bAnalysis
                try:
                    # for each spike in spikeList
                    # need to find the index into the actual plot
                    # this is the only place we need index() and it runs at n^2
                    _plotSpikeIndex = self._plotSpikeNumber.index(_realSpikeNumber)
                    _plotSpikeIndexList.append(_plotSpikeIndex)
                except (ValueError) as e:
                    logger.warning(f'we are not plotting spike number {_realSpikeNumber}')

            # xData = [self.xData[spikeList]]
            # yData = [self.yData[spikeList]]
            xData = [self.xData[_plotSpikeIndexList]]
            yData = [self.yData[_plotSpikeIndexList]]

        else:
            xData = []
            yData = []

        # logger.info(f'!!! SET DATA {xData} {yData}')
        #self.spikeListSel.set_data(xData, yData)
        #self.myHighlighter.setData(xData, yData, _plotSpikeIndexList)
        self.myHighlighter.selectSpikeList(_plotSpikeIndexList)

        # update a little info about first spike
        _str = ""
        if len(spikeList) > 0:
            firstSpike = spikeList[0]
            _oneDict = self.ba.getOneSpikeDict(firstSpike)
            _str += f"Number {_oneDict['spikeNumber']}"
            _str += f" UserType {_oneDict['userType']}"
            _str += f" Sweep {_oneDict['sweep']}"
            _str += f" Epoch {_oneDict['epoch']}"

        self.spikeNumberLabel.setText(f"First Spike Selection: {_str}")

        self.static_canvas.draw()
        # was this
        # self.repaint() # update the widget

    def old_selectSpike(self, sDict=None):
        """Select a spike

        sDict (dict): NOT USED
        """

        # logger.info(sDict)

        # spikeNumber = sDict['spikeNumber']
        # spikeList = [spikeNumber]

        spikeNumber = self.selectedSpike

        if self.xStatName is None or self.yStatName is None:
            return

        xData = []
        yData = []
        if spikeNumber is not None:
            try:
                xData = [self.xData[spikeNumber]]
                yData = [self.yData[spikeNumber]]
            except IndexError as e:
                logger.error(e)

        self.spikeSel.set_data(xData, yData)
        # self.spikeSel.set_offsets(xData, yData)

        self.spikeNumberLabel.setText(f"Spike: {spikeNumber}")

        self.static_canvas.draw()
        # was this
        # self.repaint() # update the widget

    def selectSpikesFromHighlighter(self, selectedSpikeList):
        """User selected some spikes with Highlighter -->> Propogate to main.
        
        Notes
        =====
        selectedSpikeList is based on what we are plotting (a subset of all spikes)
        get actual spikes from self._plotSpikeNumber
        """
        # logger.info(f'got highlighter plot selectedSpikeList: {selectedSpikeList}')

        if len(selectedSpikeList) > 0:
            actualSpikeList = []
            for _plotSpike in selectedSpikeList:
                try:
                    actualSpike = self._plotSpikeNumber[_plotSpike]
                    actualSpikeList.append(actualSpike)
                except (IndexError) as e:
                    logger.warning(f'we are not plotting spike {_plotSpike}')

            # print('    selectedSpikeList:', selectedSpikeList)
            # print('    actualSpikeList:', actualSpikeList)
            
            #self.setSelectedSpikes(selectedSpikeList)
            self.setSelectedSpikes(actualSpikeList)
            sDict = {
                "spikeList": self.getSelectedSpikes(),
                "doZoom": False,  # never zoom on multiple spike selection
                "ba": self.ba,
            }
            logger.info(f'{self._myClassName()} -->> emit signalSelectSpikeList')
            logger.info(f"spikeList: {sDict['spikeList']}")
            
            self._blockSlots = True
            self.signalSelectSpikeList.emit(sDict)
            self._blockSlots = False

    def toolbarHasSelection(self):
        """Return true if either ['zoom rect', 'pan/zoom'] are selected

        This is needed to cancel mouse clicks in Highlighter and right click in SanPyPlugin
        """
        state = self.mplToolbar.mode
        return state in ["zoom rect", "pan/zoom"]

    def prependMenus(self, contextMenu):
        """Prepend menus to context menu.
        
        Inherited from sanpyPlugin
        """

        noSpikesSelected = len(self.getSelectedSpikes()) == 0

        includeAction = contextMenu.addAction("Accept")
        includeAction.setDisabled(noSpikesSelected)  # disable if no spikes

        rejectAction = contextMenu.addAction("Reject")
        rejectAction.setDisabled(noSpikesSelected)

        userType1_action = contextMenu.addAction("User Type 1")
        userType1_action.setDisabled(noSpikesSelected)

        userType2_action = contextMenu.addAction("User Type 2")
        userType2_action.setDisabled(noSpikesSelected)

        userType3_action = contextMenu.addAction("User Type 3")
        userType3_action.setDisabled(noSpikesSelected)

        userType4_action = contextMenu.addAction("User Type 4")
        userType4_action.setDisabled(noSpikesSelected)

        userType5_action = contextMenu.addAction("User Type 5")
        userType5_action.setDisabled(noSpikesSelected)

        userType4_action = contextMenu.addAction("Reset User Type")
        userType4_action.setDisabled(noSpikesSelected)

        contextMenu.addSeparator()

    def handleContextMenu(self, action):
        """
        Returns:
            True if handled, otherwise False
        """
        if action is None:
            return False

        text = action.text()
        logger.info(f'Action text "{text}"')

        _selectedSpikes = self.getSelectedSpikes()

        # eventDict = setSpikeStatEvent
        setSpikeStatEvent = {}
        setSpikeStatEvent["ba"] = self.ba
        setSpikeStatEvent["spikeList"] = self.getSelectedSpikes()
        setSpikeStatEvent["colStr"] = None  # colStr
        setSpikeStatEvent["value"] = None  # value

        # logger.info(f"  -->> emit setSpikeStatEvent:{setSpikeStatEvent}")
        # self.signalSetSpikeStat.emit(setSpikeStatEvent)

        handled = False
        if text == "Accept":
            # print('Set selected spikes to include (not isbad)')
            # self.ba.setSpikeStat(_selectedSpikes, "include", False)
            setSpikeStatEvent["colStr"] = 'include'  # colStr
            setSpikeStatEvent["value"] = True  # value
            # self.replot()
            handled = True
        elif text == "Reject":
            # print('Set selected spikes to reject (isbad)')
            # self.ba.setSpikeStat(_selectedSpikes, "include", True)
            setSpikeStatEvent["colStr"] = 'include'  # colStr
            setSpikeStatEvent["value"] = False  # value
            # self.replot()
            handled = True
        elif text == "User Type 1":
            # self.ba.setSpikeStat(_selectedSpikes, "userType", 1)
            setSpikeStatEvent["colStr"] = 'userType'  # colStr
            setSpikeStatEvent["value"] = 1  # value
            # self.replot()
            handled = True
        elif text == "User Type 2":
            # self.ba.setSpikeStat(_selectedSpikes, "userType", 2)
            setSpikeStatEvent["colStr"] = 'userType'  # colStr
            setSpikeStatEvent["value"] = 2  # value
            # self.replot()
            handled = True
        elif text == "User Type 3":
            # self.ba.setSpikeStat(_selectedSpikes, "userType", 3)
            setSpikeStatEvent["colStr"] = 'userType'  # colStr
            setSpikeStatEvent["value"] = 3  # value
            # self.replot()
            handled = True
        elif text == "User Type 4":
            setSpikeStatEvent["colStr"] = 'userType'  # colStr
            setSpikeStatEvent["value"] = 4  # value
            handled = True
        elif text == "User Type 5":
            setSpikeStatEvent["colStr"] = 'userType'  # colStr
            setSpikeStatEvent["value"] = 5  # value
            handled = True
        elif text == "Reset User Type":
            # self.ba.setSpikeStat(_selectedSpikes, "userType", 0)
            setSpikeStatEvent["colStr"] = 'userType'  # colStr
            setSpikeStatEvent["value"] = 0  # value
            # self.replot()
            handled = True
        else:
            # logger.info(f'Action not understood "{text}"')
            pass

        if handled:
            logger.info(f"  -->> emit setSpikeStatEvent:{setSpikeStatEvent}")
            self.signalUpdateAnalysis.emit(setSpikeStatEvent)

        return handled

    def copyToClipboard(self):
        """Copy current x/y stats to clipboard with some other book-keeping.

        For example: spike number, spike time (s), is bad, user type, and file name.
        """
        spikeNumber = self.getStat("spikeNumber")
        spikeTimeSec = self.getStat("thresholdSec")
        xStat = self.getStat(self.xStatName)
        yStat = self.getStat(self.yStatName)
        goodSpikes = self.getStat("include")
        userType = self.getStat("userType")
        file = self.getStat("file")

        columns = [
            "Spike Number",
            "Spike Time(s)",
            self.xStatHumanName,
            self.yStatHumanName,
            "Include",
            "User Type",
            "File",
        ]
        df = pd.DataFrame(columns=columns)
        df["Spike Number"] = spikeNumber
        df["Spike Time(s)"] = spikeTimeSec
        df[self.xStatHumanName] = xStat
        df[self.yStatHumanName] = yStat
        df["Include"] = goodSpikes
        df["User Type"] = userType
        df["File"] = file

        excel = True
        sep = "\t"
        df.to_clipboard(excel=excel, sep=sep)

        logger.info(f"Copied {len(df)} spikes to clipboard.")


class myStatListWidget(QtWidgets.QWidget):
    """
    Widget to display a table with selectable stats.

    Gets list of stats from: sanpy.bAnalysisUtil.getStatList()
    """

    def __init__(self, myParent, statList=None, headerStr="Stat", parent=None):
        """
        Parameters
        ----------
        myParent : sanpy.interface.plugins.sanpyPlugin
        """
        super().__init__(parent)

        self.myParent = myParent
        if statList is not None:
            self.statList = statList
        else:
            self.statList = sanpy.bAnalysisUtil.getStatList()
        self._rowHeight = 9

        self.myQVBoxLayout = QtWidgets.QVBoxLayout(self)

        self.myTableWidget = QtWidgets.QTableWidget()
        self.myTableWidget.setWordWrap(False)
        self.myTableWidget.setRowCount(len(self.statList))
        self.myTableWidget.setColumnCount(1)
        self.myTableWidget.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.myTableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.myTableWidget.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.myTableWidget.cellClicked.connect(self.on_scatter_toolbar_table_click)

        # set font size of table (default seems to be 13 point)
        # fnt = self.font()
        # fnt.setPointSize(self._rowHeight)
        # self.myTableWidget.setFont(fnt)

        headerLabels = [headerStr]
        self.myTableWidget.setHorizontalHeaderLabels(headerLabels)

        header = self.myTableWidget.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        # QHeaderView will automatically resize the section to fill the available space. The size cannot be changed by the user or programmatically.
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)

        for idx, stat in enumerate(self.statList):
            item = QtWidgets.QTableWidgetItem(stat)
            self.myTableWidget.setItem(idx, 0, item)
            self.myTableWidget.setRowHeight(idx, self._rowHeight)

        # assuming dark theme
        # does not work
        """
        p = self.myTableWidget.palette()
        color1 = QtGui.QColor('#222222')
        color2 = QtGui.QColor('#555555')
        p.setColor(QtGui.QPalette.Base, color1)
        p.setColor(QtGui.QPalette.AlternateBase, color2)
        self.myTableWidget.setPalette(p)
        self.myTableWidget.setAlternatingRowColors(True)
        """
        self.myQVBoxLayout.addWidget(self.myTableWidget)

        # select a default stat
        self.myTableWidget.selectRow(0)  # hard coding 'Spike Frequency (Hz)'

    def getCurrentRow(self):
        return self.myTableWidget.currentRow()

    def getCurrentStat(self):
        # assuming single selection
        row = self.getCurrentRow()
        humanStat = self.myTableWidget.item(row, 0).text()

        # convert from human readbale to backend
        try:
            stat = self.statList[humanStat]["name"]
        except KeyError as e:
            logger.error(f'Did not find humanStat "{humanStat}" exception:{e}')
            humanStat = None
            stat = None
            # for k,v in

        return humanStat, stat

    @QtCore.pyqtSlot()
    def on_scatter_toolbar_table_click(self):
        """
        replot the stat based on selected row
        """
        # print('*** on table click ***')
        row = self.myTableWidget.currentRow()
        if row == -1 or row is None:
            return
        # yStat = self.myTableWidget.item(row,0).text()
        self.myParent.replot()

    """
    @QtCore.pyqtSlot()
    def on_button_click(self, name):
        print('=== myStatPlotToolbarWidget.on_button_click() name:', name)
    """


class Highlighter(object):
    """
    See: https://stackoverflow.com/questions/31919765/choosing-a-box-of-data-points-from-a-plot
    """

    def __init__(self, parentPlot, ax, x, y, plotSpikeNumber):
        self._parentPlot = parentPlot
        self.ax = ax
        self.canvas = ax.figure.canvas

        self.x = None  # these are set in setData
        self.y = None
        self._plotSpikeNumber = None
        #self._plotSpikeNumber = None

        self.setData(x, y, self._plotSpikeNumber)

        # mask will be set in self.setData
        if x and y:
            self.mask = np.zeros(x.shape, dtype=bool)
        else:
            self.mask = None

        markerSize = 50
        self._highlight = ax.scatter([], [], s=markerSize, color="yellow", zorder=10)

        # here self is setting the callback and calls __call__
        # self.selector = RectangleSelector(ax, self, useblit=True, interactive=False)
        # matplotlib.widgets.RectangleSelector
        self.selector = RectangleSelector(
            ax,
            self._HighlighterReleasedEvent,
            button=[1],
            useblit=True,
            interactive=False,
        )

        self.mouseDownEvent = None
        self.keyIsDown = None

        # april 2023, adding ?
        self._keepPickEvent = self.ax.figure.canvas.mpl_connect("pick_event", self._on_spike_pick_event3)

        self.ax.figure.canvas.mpl_connect("key_press_event", self._keyPressEvent)
        self.ax.figure.canvas.mpl_connect("key_release_event", self._keyReleaseEvent)

        # remember, sanpyPlugin is installing for key press and on pick
        self.keepOnMotion = self.ax.figure.canvas.mpl_connect(
            "motion_notify_event", self.on_mouse_move
        )
        self.keepMouseDown = self.ax.figure.canvas.mpl_connect(
            "button_press_event", self.on_button_press
        )
        self._keepMouseDown = self.ax.figure.canvas.mpl_connect(
            "button_release_event", self.on_button_release
        )

    def _keyPressEvent(self, event):
        # logger.info(event)
        self.keyIsDown = event.key

    def _keyReleaseEvent(self, event):
        # logger.info(event)
        self.keyIsDown = None

    def _on_spike_pick_event3(self, event):
        """
        
        Parameters
        ----------
        event : matplotlib.backend_bases.PickEvent
        """

        # ignore when not left mouse button
        if event.mouseevent.button != 1:
            return

        # no hits
        if len(event.ind) < 1:
            return

        _clickedPlotIdx = event.ind[0]
        logger.info(f'HighLighter _clickedPlotIdx: {_clickedPlotIdx} keyIsDown:{self.keyIsDown}')

        # convert to what we are actually plotting
        try:
            #_realSpikeNumber = self._plotSpikeNumber.index(_clickedPlotIdx)
            # get real spike number from subset of plotted psikes
            _realSpikeNumber = self._plotSpikeNumber[_clickedPlotIdx]
        except (IndexError) as e:
            logger.warning(f'  xxx we are not plotting _clickedPlotIdx {_clickedPlotIdx}')

        logger.info(f'_realSpikeNumber: {_realSpikeNumber}')

        # xData = self.x[_realSpikeNumber]
        # yData = self.y[_realSpikeNumber]
        # self._highlight.set_offsets([xData, yData])

        # if shift then add to mask
        # self.mask |= _insideMask
        newMask = np.zeros(self.x.shape, dtype=bool)
        newMask[_clickedPlotIdx] = True
        
        if self.keyIsDown == "shift":
            # oldMask = np.where(self.mask == True)
            # oldMask = oldMask[0]  # why does np do this ???
            # print('oldMask:', oldMask)

            newSelectedSpikes = np.where(newMask == True)
            newSelectedSpikes = newSelectedSpikes[0]  # why does np do this ???
            #print('newSelectedSpikes:', newSelectedSpikes)

            # add to mask
            self.mask |= newMask

            # print('newMask:')
            # print(newMask)
            # print('self.mask:')
            # print(self.mask)
            
        else:
            # replace with new
            self.mask = newMask

        # newMask = np.where(self.mask == True)
        # newMask = newMask[0]  # why does np do this ???
        # print('2) newMask:', newMask)

        xy = np.column_stack([self.x[self.mask], self.y[self.mask]])
        self._highlight.set_offsets(xy)
        
        self._HighlighterReleasedEvent()

        self.canvas.draw()

    def on_button_press(self, event):
        """
        Args:
            event : matplotlib.backend_bases.MouseEvent
        """
    
        # logger.info(f'Highlighter')

        # don't take action on right-click
        if event.button != 1:
            # not the left button
            # print('  rejecting not button 1')
            return

        # do nothing in zoom or pan/zoom is active
        # finding documentation on mpl toolbar is near impossible
        # https://stackoverflow.com/questions/20711148/ignore-matplotlib-cursor-widget-when-toolbar-widget-selected
        # state = self._parentPlot.static_canvas.manager.toolbar.mode  # manager is coming up None
        if self._parentPlot.toolbarHasSelection():
            return
        # was this
        # state = self._parentPlot.mplToolbar.mode
        # if state in ['zoom rect', 'pan/zoom']:
        #    logger.info(f'Ignoring because tool "{state}" is active')
        #    return

        self.mouseDownEvent = event
    
        # not sure why I was clearing the mask here
        if self.keyIsDown == "shift":
            # if shift is down then add to mask
            pass
        else:
            # create a new mask
            #logger.info('CLEARING MASK')
            #self.mask = np.zeros(self.x.shape, dtype=bool)
            pass

    def on_button_release(self, event):
        logger.info(f'Highlighter')

        # don't take action on right-click
        if event.button != 1:
            # not the left button
            # print('  rejecting not button 1')
            return

        self.mouseDownEvent = None

    def on_mouse_move(self, event):
        """When mouse is down, respond to movement and select points.

        Parameters
        ----------
        event : matplotlib.backend_bases.MouseEvent

        Notes
        -----
        event contains:
            motion_notify_event: xy=(113, 36)
            xydata=(None, None)
            button=None
            dblclick=False
            inaxes=None
        """

        # self.ax is our main scatter plot axes
        if event.inaxes != self.ax:
            return

        # mouse is not down
        if self.mouseDownEvent is None:
            return

        event1 = self.mouseDownEvent
        event2 = event

        if event1 is None or event2 is None:
            return

        _insideMask = self.inside(event1, event2)
        # print(f'_insideMask: {_insideMask}')

        self.mask |= _insideMask
        xy = np.column_stack([self.x[self.mask], self.y[self.mask]])
        self._highlight.set_offsets(xy)
        self.canvas.draw()

    def setData(self, x, y, plotSpikeNumber : List[int]):
        """Set underlying highlighter data, call this when we replot() scatter
        
        """
        # convert list to np array
        xArray = np.array(x)
        yArray = np.array(y)

        self.mask = np.zeros(xArray.shape, dtype=bool)
        self.x = xArray
        self.y = yArray
        self._plotSpikeNumber = plotSpikeNumber

    def selectSpikeList(self, plotIdxList):
        """
        plotIdxList : List[int]
            List of plot indices to select
        """
        self.mask = np.zeros(self.x.shape, dtype=bool)
        self.mask[plotIdxList] = True

        xy = np.column_stack([self.x[self.mask], self.y[self.mask]])
        self._highlight.set_offsets(xy)
        
        # self._HighlighterReleasedEvent()

        self.canvas.draw()

    # def selectSpikes(self, spikeList):
        
    #     self.mask = np.zeros(self.x.shape, dtype=bool)
        
    #     # the spike numbers we are plotting self._plotSpikeNumber
    #     _selectionIdx = []
    #     for spike in spikeList:
    #         if spike in self._plotSpikeNumber:

    #     self.mask

    #     xy = np.column_stack([self.x[self.mask], self.y[self.mask]])
    #     self._highlight.set_offsets(xy)
        
    #     self._HighlighterReleasedEvent()

    #     self.canvas.draw()

    # def __call__(self, event1, event2):
    def _HighlighterReleasedEvent(self, event1=None, event2=None):
        """RectangleSelector callback when mouse is released

        event1:
            button_press_event: xy=(87.0, 136.99999999999991) xydata=(27.912559411227885, 538.8555851528383) button=1 dblclick=False inaxes=AxesSubplot(0.1,0.1;0.607046x0.607046)
        event2:
            button_release_event: xy=(131.0, 211.99999999999991) xydata=(48.83371692821588, 657.6677439956331) button=1 dblclick=False inaxes=AxesSubplot(0.1,0.1;0.607046x0.607046)
        """

        self.mouseDownEvent = None

        # emit the selected spikes
        selectedSpikes = np.where(self.mask == True)
        selectedSpikes = selectedSpikes[0]  # why does np do this ???

        logger.info(f'selectedSpikes: {selectedSpikes}')

        selectedSpikesList = selectedSpikes.tolist()
        self._parentPlot.selectSpikesFromHighlighter(selectedSpikesList)

        # we now use self._blockSlots
        # # clear the selection user just made, will get 'reselected' in signal/slot
        #self._highlight.set_offsets([np.nan, np.nan])

        return

    def inside(self, event1, event2):
        """Returns a boolean mask of the points inside the
        rectangle defined by event1 and event2.
        """
        # Note: Could use points_inside_poly, as well
        x0, x1 = sorted([event1.xdata, event2.xdata])
        y0, y1 = sorted([event1.ydata, event2.ydata])
        mask = (self.x > x0) & (self.x < x1) & (self.y > y0) & (self.y < y1)
        return mask


def run():
    path = "/Users/cudmore/Sites/SanPy/data/19114001.abf"
    ba = sanpy.bAnalysis(path)
    ba.spikeDetect()
    print(ba)

    import sys

    app = QtWidgets.QApplication([])
    sp = plotScatter(ba=ba)
    sp.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    run()



import sanpy
from sanpy.interface.plugins import basePlotTool

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class plotToolPool(basePlotTool):
    """Plot tool pooled across all open analysis"""

    myHumanName = "Plot Tool (pool)"

    def __init__(self, tmpMasterDf=None, **kwargs):
        """
        tmpMasterDf (pd df): only for debuggin
        """
        super().__init__(**kwargs)

        self.masterDf = None
        if self.getSanPyApp() is not None:
            self.masterDf = self.getSanPyApp().myAnalysisDir.pool_build()
            logger.info(self.masterDf)
            self.masterDf.to_csv("/Users/cudmore/Desktop/tmpDf-20221231.csv")
        elif tmpMasterDf is not None:
            logger.info("Using tmpMasterDf")
            self.masterDf = tmpMasterDf

        self.plot()


if __name__ == "__main__":
    import sys
    from PyQt5 import QtCore, QtWidgets, QtGui

    app = QtWidgets.QApplication([])

    # load an analysis dir
    path = "/home/cudmore/Sites/SanPy/data"
    ad = sanpy.analysisDir(path, autoLoad=True)
    print(ad._df)
    # load all analysis
    print("\n=== loading all analysis")
    for fileIdx in range(ad.numFiles):
        ad.getAnalysis(fileIdx)
    # build the pool
    print("\n=== building pool")
    tmpMasterDf = ad.pool_build()

    # open window
    print("\n=== opening plotToolPool with tmpMasterDf:", len(tmpMasterDf))
    # print('tmpMasterDf.columns:', tmpMasterDf.columns)
    ptp = plotToolPool(tmpMasterDf=tmpMasterDf)
    ptp.show()

    sys.exit(app.exec_())



from typing import Union, Dict, List, Tuple
import numpy as np
import pyabf

import sanpy
from sanpy.fileloaders.fileLoader_base import fileLoader_base, recordingModes

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class fileLoader_abf(fileLoader_base):
    loadFileType = "abf"

    # @property
    # def loadFileType(self):
    #     return 'abf'

    def loadFile(self):
        self._loadAbf()

    def _loadAbf(
        self, byteStream=None, loadData: bool = True, stimulusFileFolder: str = None
    ):
        """Load pyAbf from path."""
        try:
            # logger.info(f'loadData:{loadData}')
            if byteStream is not None:
                logger.info("Loading byte stream")
                self._abf = pyabf.ABF(byteStream)
                self._isBytesIO = True
            else:
                # logger.info(f'Loading file: {self.filepath}')
                self._abf = pyabf.ABF(
                    self.filepath,
                    loadData=loadData,
                    stimulusFileFolder=stimulusFileFolder,
                )

        except NotImplementedError as e:
            logger.error(f"    did not load abf file: {self.filepath}")
            logger.error(f"      NotImplementedError exception was: {e}")
            self._loadError = True
            self._abf = None

        except Exception as e:
            # some abf files throw: 'unpack requires a buffer of 234 bytes'
            # 'ABF' object has no attribute 'sweepEpochs'
            logger.error(f"    did not load abf file: {self.filepath}")
            logger.error(f"        unknown Exception was: {e}")
            self._loadError = True
            self._abf = None

        if loadData:
            try:
                _tmp = self._abf.sweepEpochs.p1s
            except AttributeError as e:
                logger.warning(
                    f"    did not find epochTable loadData:{loadData}: {e} in file {self.filepath}"
                )
            else:
                _numSweeps = len(self._abf.sweepList)
                self._epochTableList = [None] * _numSweeps
                for _sweepIdx in range(_numSweeps):
                    self._abf.setSweep(_sweepIdx)
                    self._epochTableList[_sweepIdx] = sanpy.fileloaders.epochTable(
                        self._abf
                    )
                self._abf.setSweep(0)

            self._sweepList = self._abf.sweepList
            self._sweepLengthSec = (
                self._abf.sweepLengthSec
            )  # assuming all sweeps have the same duration

            # on load, sweep is 0
            if loadData:
                _numRows = self._abf.sweepX.shape[0]
                numSweeps = len(self._sweepList)
                self._sweepX = np.zeros((_numRows, 1))
                self._sweepY = np.zeros((_numRows, numSweeps))
                self._sweepC = np.zeros((_numRows, numSweeps))

                _channel = 0
                for sweep in self._sweepList:
                    self._abf.setSweep(sweepNumber=sweep, channel=_channel)
                    if sweep == 0:
                        self._sweepX[
                            :, sweep
                        ] = self._abf.sweepX  # <class 'numpy.ndarray'>, (60000,)
                    self._sweepY[:, sweep] = self._abf.sweepY
                    try:
                        self._sweepC[:, sweep] = self._abf.sweepC
                    except ValueError as e:
                        # pyabf will raise this error if it is an atf file
                        logger.warning(
                            f"    exception fetching sweepC for sweep {sweep} with {self.numSweeps}: {e}"
                        )
                        #
                        # if we were recorded with a stimulus file abf
                        # needed to assign stimulusWaveformFromFile
                        try:
                            tmpSweepC = self._abf.sweepC
                            self._sweepC[:, sweep] = self._abf.sweepC
                        except ValueError as e:
                            logger.warning(f"ba has no sweep {sweep} sweepC ???")

                # not needed
                self._abf.setSweep(0)

            # get v from pyAbf
            self._dataPointsPerMs = self._abf.dataPointsPerMs

            # turned back on when implementing Santana rabbit Ca kymographs
            abfDateTime = self._abf.abfDateTime  # 2019-01-14 15:20:48.196000
            self._acqDate = abfDateTime.strftime("%Y-%m-%d")
            self._acqTime = abfDateTime.strftime("%H:%M:%S")

            self._numChannels = len(self._abf.adcUnits)
            if self._numChannels > 1:
                logger.warning(
                    f"    SanPy does not work with multi-channel recordings numChannels is {self._numChannels} {self._path}"
                )
                # logger.warning('    Will default to channel 0')

            # self.sweepUnitsY = self.adcUnits[channel]
            channel = 0
            # dacUnits = self._abf.dacUnits[channel]
            adcUnits = self._abf.adcUnits[channel]
            # print('  adcUnits:', adcUnits)  # 'mV'
            self._sweepLabelY = adcUnits
            self._sweepLabelX = "sec"

            # self._sweepLabelX = self._abf.sweepLabelX
            # self._sweepLabelY = self._abf.sweepLabelY
            if self._sweepLabelY in ["pA", "nA"]:
                self._recordingMode = recordingModes.vclamp  # 'V-Clamp'
                # self._sweepY_label = self._abf.sweepUnitsY
            elif self._sweepLabelY in ["mV"]:
                self._recordingMode = recordingModes.iclamp  #'I-Clamp'
                # self._sweepY_label = self._abf.sweepUnitsY
            else:
                logger.warning(f'did not understand adcUnit "{adcUnits}"')

        #
        self.myFileType = "abf"

        # base sanpy does not keep the abf around
        # logger.warning('[[[TURNED BACK ON]]] I turned off assigning self._abf=None for stoch-res stim file load')
        self._abf = None



"""
Started versioning very late, as of 20210425

analysisVersion: saved for each spike
				indicates the analysis we used
				update this file as I change the analysis
"""

# 20210425
# analysisVersion = '20210426a' # backend engine
# interfaceVersion = '20210426a' # pyqt interface

# Laura returned from Italy, getting her most recent version
# analysisVersion = '20210803a' # backend engine
analysisVersion = "20230324a"  # backend engine

interfaceVersion = "20230324a"  # pyqt interface

# manually keep in sync with setup.py
__version__ = "0.1.11"